
/*
 * The Verus Model Checking System
 * Copyright (c) 1997 Carnegie Mellon University
 * All Rights Reserved. See file license.txt for details.
 *
 *
 * Routines for manipulating BDDs: redirects bdds operations to proper
 * routine in the bdd package.
 *
 * Sergio Campos -- 1997
 *
 */

#include<stdlib.h>
#include<strings.h>
#include<bdd.h>
#include<symbols.h>
#include<integer.h>
#include<compile.h>


#define BDD_IS_CONST(b) ((b == bdd_one(bddm)) || (b == bdd_zero(bddm)))

/*
 * Used to keep a list of bdds to be freed. Bdds are not freed when
 * v_free_bdd is called, the last cache_size bdds are kept around.
 * This is done because bdds created later might have subexpressions
 * in common with previous bdds, and if those are kept around for
 * some time longer, they might be reused.
 *
 * Another fact that affects the size and time of the verification is the
 * frequency of garbage collect calls. This is controlled by the gc_max
 * variable. Only once every gc_max calls to bdd_free do we garbage collect.
 * This ensures that the most recent bdds created are kept around longer,
 * and can significantly speed up the verification.
 *
 * Both methods, controlled by cache_size and gc_max, are somewhat
 * redundant. The frequency of garbage collection seems to have a more
 * profound effect, so that's the one changed with a run-time option. The
 * use of a very small bdd cache in Verus serves mostly to be sure that
 * some bdds get to be reused even when a garbage collection occurs,
 * because the last cache_size bdds haven't been freed yet at gc time.
 *
 * Note that the more bdds reused, the faster things go. However, more
 * space might be wasted as well. A compromise is reached by fine-tuning
 * the value of gc_max and (to less extent) cache_size.
 */
int cache_size = 500;
bdd *cache;
int cache_next = 0;
double max_bdds_used = 0;
int gc_max = 500;
int gc_count = 0;


extern int    nvars;   /* total number of boolean variables in the program */
extern int   gnvars;   /* total number of global variables in the program */
extern int  inside_function; /* differentiates between local and global vars */


/*
 * Variable associations for model checking functions. There are two
 * of them. One for current state variables (associated with next
 * state variables), and one for next state variables (associated
 * with current state variables)
 *
 * They are used whenever step_[forward,backwards] are called, and
 * serve to existentially quantify out variables as well as for
 * shifting variables to/from current/next state.
 */
bdd current_assoc_array[V_MAX_VARIABLES + 1];
bdd next_assoc_array[V_MAX_VARIABLES + 1];
int assoc_last;
int current_assoc;
int next_assoc;


/*
 * User variables are kept in a separate association, to be printed
 * independently of waitcounters.
 */
bdd uservars_assoc_array[V_MAX_VARIABLES + 1];
int uservars_assoc;
int uservars_last;


void init_bdds()
{
  int i;

  bddm = bdd_init();
  printf("Using BDD library version %s.\n", bdd_version());

  /* initialize bdd cache */
  cache = malloc(cache_size * sizeof(bdd));
  for (i = 0; i < cache_size; i++) cache[i] = bdd_zero(bddm);
}


void init_assoc()
{
  for(assoc_last = 0; assoc_last <= V_MAX_VARIABLES; assoc_last++) {
    current_assoc_array[assoc_last] = 0;
    next_assoc_array[assoc_last] = 0;
    uservars_assoc_array[assoc_last] = 0;
  };
  assoc_last = 0;
  uservars_last = 0;
};


void setup_assocs()
{
  current_assoc = bdd_new_assoc(bddm, current_assoc_array, 1);
  next_assoc = bdd_new_assoc(bddm, next_assoc_array, 1);
  uservars_assoc = bdd_new_assoc(bddm, uservars_assoc_array, 0);
};


/*
 * creates a new bdd for the variable of type vtype and returns its id.
 */
int v_new_bdd(symbol_name, vtype, v1, v2, v3)
char *symbol_name;
variable_type vtype;
int *v1;
int *v2;
int *v3;
{
  int result;
  int pos;       /* index of variable as defined in ordering file */
  char *s;

  switch (vtype) {
    case VARTYPE_BOOLEAN:
                  pos = read_order_file_lookup(symbol_name);
                  if (pos != VNOTFOUND) {
                    /* variables have already been created. Find out
                       who they are */
                    (*v2) = (int) bdd_var_with_index(bddm, pos); /* current */
                    s = malloc(strlen(symbol_name) + 5);
                    strcpy(s, symbol_name);
                    strcat(s, "'");
                    pos = read_order_file_lookup(s);
                    if (pos == VNOTFOUND)
                      fatal_error("Missing variable in ordering file: %s", s);
                    (*v1) = (int) bdd_var_with_index(bddm, pos); /* next    */

                    strcpy(s, symbol_name);
                    strcat(s, "_aux");
                    pos = read_order_file_lookup(s);
                    if (pos == VNOTFOUND)
                      fatal_error("Missing variable in ordering file: %s", s);
                    (*v3) = (int) bdd_var_with_index(bddm, pos); /* aux     */
                    free(s);
                  } else {
                    if (varname_lastvar != 0)
                      error("Creating variable not in order file: %s.", 
                            symbol_name);
                    (*v3) = (int) bdd_new_var_last(bddm); /* aux     */
                    (*v2) = (int) bdd_new_var_last(bddm); /* current */
                    (*v1) = (int) bdd_new_var_last(bddm); /* next    */
                  };
                  if (assoc_last+2 > V_MAX_VARIABLES) {
                    fatal_error("Maximum number of variables exceeded.\n");
                  };
                  current_assoc_array[assoc_last] = (bdd) *v2;
                  next_assoc_array[assoc_last]    = (bdd) *v1;
                  assoc_last++;
                  current_assoc_array[assoc_last] = (bdd) *v1;
                  next_assoc_array[assoc_last]    = (bdd) *v2;
                  assoc_last++;

                  if (inside_function) {
                    nvars++;   /* a new variable is born */
                  } else {
                    gnvars++;  /* a new global variable is born */
                  };

                  break;
    case VARTYPE_INTEGER:
                  new_int_var(symbol_name, DEFAULT_INT_SIZE, v1, v2, v3);
                  if (inside_function) {
                    nvars += DEFAULT_INT_SIZE;  /* many vars are born */
                  } else {
                    gnvars += DEFAULT_INT_SIZE; /* many global vars are born */
                  };
                  break;
  };

  return(result);
};


extern char *out_ord_file;
extern double statespace;


/*
 * Releases bdd space
 */
void v_free_bdd(b)
bdd b;
{
  double bdds_used;

  if (!BDD_IS_CONST(b)) {
    bdd_free(bddm, cache[cache_next]);
    cache[cache_next] = b;
    cache_next++;
    if (cache_next >= cache_size) cache_next = 0;

    gc_count++;
    if (gc_count >= gc_max) {
      gc_count = 0;
      bdds_used = bdd_total_size(bddm);
      if (bdds_used > max_bdds_used)  max_bdds_used = bdds_used;

      /* clean up old bdds... */
      bdd_gc(bddm);
    };
  };
};


/*
 * BDD stack implementation - Used to store intermediate relations
 *    during compilation.
 */

/* initialized by default to empty */
stack_ptr Rstack = NULL;

vresult Rpush(b)
bdd b;
{
  stack_ptr st;

  st = (stack_ptr) malloc(sizeof(struct stack));
  if (st == NULL) {
    fatal_error("bdd.c: Rpush: Unable to allocate Rstack.");
  };
  st->top = b;
  st->next = Rstack;

  Rstack = st;

  return(VOK);
};

bdd Rpop()
{
  bdd b;
  stack_ptr st;

  if (Rstack == NULL) {
    fatal_error("bdd.c: Rpop: empty Rstack.");
  };
  b = Rstack->top;
  st = Rstack->next;

  free(Rstack);
  Rstack = st;

  return(b);
};


/*
 * These routines pretty print bdds. To do this we must keep a table
 * from bdds to variable names. The routine insert_symbol makes an
 * entry in this table. The routine v_print_bdd does the nice work.
 */
struct bdd_to_name_ttype{
  char           *name;
  bdd            b;
};
#define bdd_to_name_MAX (2 * V_MAX_VARIABLES)
struct bdd_to_name_ttype bdd_to_name_table[bdd_to_name_MAX];
int               bdd_to_name_last = 0;


void bdd_to_name_insert(p)
symbol_ptr(p);
{
  char *c;
  char *s;
  char si[4];
  int   i, n;
  bdd  *b;

  c = p->name;

  /* insert on table */

  if (bdd_to_name_last >= bdd_to_name_MAX) {
    error("overflow of bdd_to_name table. Variable %s not inserted", c);
    return;
  }
  if (p->vartype == VARTYPE_BOOLEAN) {
    bdd_to_name_table[bdd_to_name_last].name = c;
    bdd_to_name_table[bdd_to_name_last].b = (bdd) p->value2;
    bdd_to_name_last++;
    if (bdd_to_name_last >= bdd_to_name_MAX)
      error("overflow of bdd_to_name table.");
    s = malloc(strlen(c) + 2);
    strcpy(s, c);
    strcat(s, "'");
    bdd_to_name_table[bdd_to_name_last].name = s;
    bdd_to_name_table[bdd_to_name_last].b = (bdd) p->value1;
    bdd_to_name_last++;
    if (bdd_to_name_last >= bdd_to_name_MAX)
      error("overflow of bdd_to_name table.");

    s = malloc(strlen(c) + 5);
    strcpy(s, c);
    strcat(s, "_aux");
    bdd_to_name_table[bdd_to_name_last].name = s;
    bdd_to_name_table[bdd_to_name_last].b = (bdd) p->value3;
    bdd_to_name_last++;
    if (bdd_to_name_last >= bdd_to_name_MAX)
      error("overflow of bdd_to_name table.");

  } else {
    b = (bdd *)p->value2;
    for(n = 0; b[n] != 0; n++); /* find last var in intvar, it is the LSB */
    n--;
    for(i = n; i >= 0; i--) {
      s = malloc(strlen(c) + 10);
      strcpy(s, c);
      strcat(s, "[");
      sprintf(si, "%i", n-i);
      strcat(s, si);
      strcat(s, "]");
      bdd_to_name_table[bdd_to_name_last].name = s;
      bdd_to_name_table[bdd_to_name_last].b = (bdd) b[i];
      bdd_to_name_last++;
      if (bdd_to_name_last >= bdd_to_name_MAX)
        error("overflow of bdd_to_name table.");
    };
    /* second var has the same size, no need to recompute n */
    b = (bdd *)p->value1;
    for(i = n; i >= 0; i--) {
      s = malloc(strlen(c) + 10);
      strcpy(s, c);
      strcat(s, "'[");
      sprintf(si, "%i", n-i);
      strcat(s, si);
      strcat(s, "]");
      bdd_to_name_table[bdd_to_name_last].name = s;
      bdd_to_name_table[bdd_to_name_last].b = (bdd) b[i];
      bdd_to_name_last++;
      if (bdd_to_name_last >= bdd_to_name_MAX)
        error("overflow of bdd_to_name table.");
    };

    b = (bdd *)p->value3;
    for(i = n; i >= 0; i--) {
      s = malloc(strlen(c) + 15);
      strcpy(s, c);
      strcat(s, "_aux[");
      sprintf(si, "%i", n-i);
      strcat(s, si);
      strcat(s, "]");
      bdd_to_name_table[bdd_to_name_last].name = s;
      bdd_to_name_table[bdd_to_name_last].b = (bdd) b[i];
      bdd_to_name_last++;
      if (bdd_to_name_last >= bdd_to_name_MAX)
        error("overflow of bdd_to_name table.");
    };
  };
};


char *bdd_to_name(bddm, b, env)
bdd_manager bddm;
bdd b;
pointer env;
{
  int i;

  i = 0;
  for(i = 0; !(bdd_to_name_table[i].b == b || i == bdd_to_name_last); i++);

  if (bdd_to_name_table[i].b == b) {
    return(bdd_to_name_table[i].name);
  } else {
    return(0);
  };
};


extern int short_counterex; /* global flag to print short counterexamples. */

void v_print_name(name, value)
char *name;
int  value;
{
  char *c;
  symbol_ptr s;
  int index;
  char *st;

  if (short_counterex && !strstr(name, WAITCOUNT_NAME)) return;

  c = strchr(name, '[');
  if (c == NULL) {
    if (value) printf(" ");
          else printf("!");
    printf(name);
    printf(" ");
  } else {
    st = (char *) malloc((c - name) + 1);
    memcpy(st, name, c - name);
    st[c-name] = 0;
    sscanf(c, "[%d]", &index);
    if (lookup_symbol(st, &s) == VNOTFOUND) {
      fatal_error("Undeclared variable in bdd: %s", st);
    };
    s->current_value += value * (1<<index);
    s->nobits++;
    if (s->nobits >= DEFAULT_INT_SIZE) {
      printf("%s=%3d ", st, s->current_value);
      s->current_value = 0;
      s->nobits = 0;
    };
  };
};


void v_print_recurse(b)
bdd b;
{
  pointer env;
  bdd temp[2];

  if (BDD_IS_CONST(b)) return;

  temp[0] = bdd_else(bddm, b);
  if (temp[0] != bdd_zero(bddm)) {
    temp[1] = bdd_if(bddm, b);
    v_print_name(bdd_to_name(bddm, bdd_if(bddm, b), env), 0);
    v_free_bdd(temp[1]);
    v_print_recurse(temp[0]);
    v_free_bdd(temp[0]);
  } else {
    v_free_bdd(temp[0]);
    temp[0] = bdd_then(bddm, b);
    if (temp[0] != bdd_zero(bddm)) {
      temp[1] = bdd_if(bddm, b);
      v_print_name(bdd_to_name(bddm, temp[1], env), 1);
      v_free_bdd(temp[1]);
      v_print_recurse(temp[0]);
      v_free_bdd(temp[0]);
    };
  };
};


extern bdd valid_wc;

void v_print_bdd(s, b)
char *s;
bdd   b;
{
  bdd temp[2];

  printf("%s ", s);

/*
 * Uncomment the next two lines in order to print bdds as if-then-else 
 * structures instead of a list of states.
 */
/*
  bdd_print_bdd(bddm, b, bdd_to_name, bdd_terminal_id_fn_none, 0, stdout);
  return;
*/

  if (b == bdd_zero(bddm)) {
    printf("no states.\n");
    return;
  };

  /* make sure we don't allow invalid _wc numbers */
  b = bdd_and(bddm, b, valid_wc);

  /* Choose one state from the state set and print it. */
  bdd_assoc(bddm, current_assoc);
  temp[0] = bdd_satisfy_support(bddm, b);
  v_print_recurse(temp[0]);

  printf("\n");
  /*
   * The code below prints *all* states in a state set. Usually the
   * state sets used by this routine only contain one state anyway,
   * so we simplify it and print only one state from a given state
   * set. The code below has not been deleted because it may be
   * useful later on.
   */
  /*
   *  while (b != bdd_zero(bddm)) {
   *    printf("%3d: ", i++);
   *    bdd_assoc(bddm, current_assoc);
   *    temp[0] = bdd_satisfy_support(bddm, b);
   *    v_print_recurse(temp[0]);
   *    temp[1] = bdd_not(bddm, temp[0]);
   *    v_free_bdd(temp[0]);
   *    temp[0] = b;
   *    b = bdd_and(bddm, b, temp[1]);
   *    printf("\n");
   *    v_free_bdd(temp[0]);
   *    v_free_bdd(temp[1]);
   *  };
   */
};  


vresult write_ordering_file(filename)
char *filename;
{
  FILE *f;
  char *name;
  bdd b;
  int i;
  pointer env;

  printf("write_ordering_file %s.\n", filename);

  if (!(f = fopen(filename, "w"))) {
    fatal_error("cannot open file %s", filename);
  };

  i = 0;
  while ((b = bdd_var_with_index(bddm, i)) != NULL) {
    name = bdd_to_name(bddm, b, env);
    if (name != 0) {
      fprintf(f, name);
      fprintf(f, "\n");
    };
    i++;
  }; 

  fclose(f);
  return(VOK);
};


/*
 * In order to read the ordering file we must maintain a table with variable
 * names and the index of the corresponding bdd variables (The index of a
 * bdd var is its position in the ordering). So, we read the file and store
 * the variable names in varname_buffer. Whenever looking for the index of
 * a variable, we search varname_buffer and return i when the variable name
 * found is the ith in the buffer.
 */
#define varname_buffer_MAX 10000
char varname_buffer[varname_buffer_MAX];
char *varname_lastvar = 0;


vresult read_ordering_file(filename)
char *filename;
{
  FILE *f;
  bdd b;
  int i;
  int c;

  printf("read_ordering_file %s.\n", filename);

  if (!(f = fopen(filename, "r"))) {
    fatal_error("cannot open file %s", filename);
  };

  i = 0;
  varname_lastvar = 0;
  while ((c = fgetc(f)) != EOF) {

    /* This conditional defines what's accepted in the ordering file. For
       now everything that's separated by '\n' is a var name. No comments
       in ordering file or any other fancy stuff. */
    if (c != '\n') {
      varname_buffer[i] = c;
      i++;
    } else {
      varname_buffer[i] = 0;
      i++;
      varname_lastvar = (char *) i;

      /* create a bdd variable for this variable. No need to save it,
         because we'll access it by its index (which is the position in
         the varname list). */
      b = bdd_new_var_last(bddm);
    };
  }; 
  varname_lastvar = varname_buffer + i;
  fclose(f);
  return(VOK);
};


int read_order_file_lookup(name)
char *name;
{
  char *i;
  int pos;

  /* check if there is an ordering file */
  if (varname_lastvar == 0) return(VNOTFOUND);

  pos = 0;
  i = varname_buffer;
  while (i != varname_lastvar) {
    if (!strcmp(i, name)) {
      /* found the variable */
      return(pos);
    };
    i = i + strlen(i) + 1;
    pos++;
  };
  return(VNOTFOUND);
};
