
/*
 * The Verus Model Checking System
 * Copyright (c) 1997 Carnegie Mellon University
 * All Rights Reserved. See file license.txt for details.
 *
 *
 * Compiler for the Verus tool. Each function performs a reduction from
 * the grammar defined in grammar.y
 *
 * Sergio Campos -- 1997
 */

#include<stdio.h>
#include<strings.h>
#include<ctype.h>
#include<stdlib.h>
#include<error.h>
#include<compile.h>
#include<symbols.h>
#include<bdd.h>
#include<integer.h>
#include<y.tab.h>
#include<modcheck.h>


/*
 * R is the relation between wait statements.
 */
bdd R;

/*
 * local_identity is the local identity relation. Local variables do not
 *    change value from state to state, unless assigned otherwise. In
 *    order to maintain this "default" we create the relation (v' = v)
 *    for all local variables v, and conjunct the result into
 *    local_identity. Local variables are declared without keyword extern.
 */
bdd local_identity;

/*
 * global_identity is the same as local_identify, but for global variables.
 */
bdd global_identity;

/*
 * global_nmaintain_vars contains a list of auxiliary variables used to decide
 * wether or not to keep the v=v' clause in the relations.
 */
bdd global_nmaintain_vars;

/*
 * global_assigned_assoc is an association to eliminate all gaux from R
 */
bdd *global_assigned_array;
int  global_assigned_index;
int  global_assigned_assoc;

/*
 * inside_function is true when we are parsing a function. This is used
 *    to decide between putting a (v' = v) into the local or global identity.
 */
int inside_function = 0;

extern char *filestart;
extern long filesize;
extern long filepos;

extern int nwaitvars;    /* size of wait counter                     */
extern int nvars;        /* total number of variables in the program */
extern int wait_count;   /* total number of wait statements in the program */

fn_list_ptr functions = 0;
id_list_ptr global_vars = 0;

extern int partitionTR;


/*
 * Used to determine where specs start and end for printing.
 */
long lastspecpos;
long skip_spaces_comments();

/*
 * Function parameters. Used for partitioned transition relations.
 */
id_list_ptr fn_vars = 0;

/*
 * Prepare to compile
 */
void init_compile()
{
  init_bdds();
  init_symbol_table();
  init_assoc();
  global_identity = bdd_one(bddm);
  global_nmaintain_vars = bdd_one(bddm);
  global_assigned_array = malloc(sizeof(bdd) * MAX_SYMBOL_SIZE);
  global_assigned_array[0] = 0;
  global_assigned_index = 0;

  /* Insert true and false in the symbol table */
  insert_symbol(lookup_string(insert_string("true")), SYMTYPE_CONSTANT,
                VARTYPE_BOOLEAN, bdd_one(bddm), bdd_one(bddm), (bdd) 0);
  insert_symbol(lookup_string(insert_string("false")), SYMTYPE_CONSTANT,
                VARTYPE_BOOLEAN, bdd_zero(bddm), bdd_zero(bddm), (bdd) 0);
};


/*
 * function_definition:
 *   identifier LP identifier_list RP {init_function($1, $3);}
 *                                    declaration_list compound_statement
 *     {reduce_function_definition($1, $4, 0, 0);}
 */
int init_function(fname, parameters)
char          *fname;
id_list_ptr   parameters;
{
  int wc;
  bdd temp[2];

  strncpy(function_name, fname, MAX_SYMBOL_SIZE);
  strcat(function_name, ".");
  inside_function = 1;

  init_wait_table();
  wc = find_wc(filepos);

  /* Set up initial relations */
  local_identity = bdd_one(bddm);

  /* create wait counter variable for this function */
  fn_vars = reduce_identifier_list(0, WAITCOUNT_NAME);
  reduce_declaration(INTEGER, fn_vars, 0);

  nwaitvars = DEFAULT_INT_SIZE; /* need to know the size of wait counter
                                   to compute size of state sets */
  nvars = 0;   /* waitcounter is not a program variable */

  /* initial R = (wc = 0) */
  R = int_to_bdd(wc, id_to_bdd_current(WAITCOUNT_NAME)->tbdd);

  Init = bdd_identity(bddm, R);

  /* initial R = R && (v = v') */
  temp[0] = R;
  R = bdd_and(bddm, R, local_identity);
  v_free_bdd(temp[0]);
  temp[0] = R;
  R = bdd_and(bddm, R, global_identity);
  /*  v_free_bdd(temp[0]);*/

  TR = bdd_zero(bddm);
  return(VOK);
};

/*
 * function_definition:
 *   identifier LP identifier_list RP {init_function($1, $3);}
 *                                    declaration_list compound_statement
 *     {reduce_function_definition($1, $5, 0, 0);}
 */
int reduce_function_definition(fname, par_list, decl_list, stmt)
char          *fname;
id_list_ptr   par_list;
int           decl_list;
int           stmt;
{
  fn_list_ptr fn;

  fn = (fn_list_ptr) malloc(sizeof(struct fn_list));
  fn->fname     = fname;
  fn->TR        = TR;
  fn->Init      = Init;
  fn->nwaitvars = nwaitvars;
  fn->nvars     = nvars;
  fn->waitcount = wait_count - 1;
  fn->next      = functions;
  fn->vars      = fn_vars;
  functions     = fn;

  inside_function = 0;
  printf("\nCompiled function %s\n", function_name);
  return(VOK);
};


/*
 * identifier_list:
 *   null                             {$$ = 0;}  |
 *   identifier                       {$$ = reduce_identifier_list(0, $1);} |
 *   identifier_list COMMA identifier {$$ = reduce_identifier_list($1, $3);}
 */
id_list_ptr reduce_identifier_list(oldlist, s)
id_list_ptr oldlist;
char *s;
{
  id_list_ptr idlist;

  if (s == 0) fatal_error("empty identifier list.");

  idlist = (id_list_ptr) malloc(sizeof(struct id_list));
  idlist->element = s;
  idlist->next = oldlist;

  return(idlist);  
}


/*
 * declaration:
 *   type_specifier identifier_list SEMICOLON
 *     {reduce_declaration($1, $2);}
 */
int reduce_declaration(vartype, idlist, is_extern)
int         vartype;
id_list_ptr idlist;
int         is_extern;
{
  id_list_ptr p, paux;
  variable_type vtype;
  symbol_ptr sptr;
  bdd v1, v2, v3;
  int is_uservar;

  p = idlist;
  switch (vartype) {
    case BOOLEAN: vtype = VARTYPE_BOOLEAN;
                  break;
    case INTEGER: vtype = VARTYPE_INTEGER;
                  break;
    default:      fatal_error("unknown variable type in declaration.");
  };
  if (is_extern && !inside_function)
    fatal_error("Extern variables must be global.");

  while (p != 0) {
    char symbol_name[MAX_SYMBOL_SIZE];

    if (*function_name == 0) {
      strcpy(symbol_name, "");
    } else {
      strcpy(symbol_name, function_name);
    };
    strcat(symbol_name, p->element);
    is_uservar = strcmp(p->element, WAITCOUNT_NAME);

    if (lookup_symbol(symbol_name, &sptr) == VNOTFOUND) {
      bdd ident, temp[4];

      v_new_bdd(symbol_name, vtype, &v1, &v2, &v3);

      if (is_extern) {
        sptr =
         insert_symbol(symbol_name, SYMTYPE_EXT_VARIABLE, vtype, v1, v2, v3);
      } else {
        if (inside_function) {
          /* some local variable */
          sptr =
           insert_symbol(symbol_name, SYMTYPE_LCL_VARIABLE, vtype, v1, v2, v3);
        } else {
          /* global variable */
          sptr =
           insert_symbol(symbol_name, SYMTYPE_GBL_VARIABLE, vtype, v1, v2, v3);
        };
      };

      /* ident = ident & (v' = v) */
      if (vtype == VARTYPE_BOOLEAN) {
        ident = bdd_xnor(bddm, v1, v2);

        if (is_uservar) {
          uservars_assoc_array[uservars_last] = v2;
          if (++uservars_last == V_MAX_VARIABLES) 
                   fatal_error("overflow of user variables association");
        };
      } else {
        /* this should be in integer.c!!! */
        int i = 0;
        intvar intv1 = (intvar) v1;
        intvar intv2 = (intvar) v2;

        ident = bdd_one(bddm);
        while(intv1[i] != 0) {
          temp[0] = bdd_xnor(bddm, intv1[i], intv2[i]);
          temp[1] = ident;
          ident = bdd_and(bddm, ident, temp[0]);

          if (is_uservar) {
            uservars_assoc_array[uservars_last] = intv2[i];
            if (++uservars_last == V_MAX_VARIABLES) 
                     fatal_error("overflow of user variables association");
          };

          v_free_bdd(temp[0]);
          v_free_bdd(temp[1]);
          i++;
        };
      };
      if (!is_extern) {
        /*
         * Local variables (is_extern = 0) must update the local_identity
         * relation, to guarantee that they do not change value
         * spontaneously like extern variables.
         */
        if (inside_function) {
          /* local variable */
          temp[0] = local_identity;
          local_identity = bdd_and(bddm, local_identity, ident);
          v_free_bdd(temp[0]);
        } else {
          /* global variable */
          /* set up ident = (v = v') | v_aux <- v_aux is used on assigns */
          if (vtype == VARTYPE_BOOLEAN) {
            temp[1] = v3;                     /* nmaintain_v variable */
            temp[0] = bdd_new_var_last(bddm); /* assigned_v variable */
          } else {
            intvar intv3;

            intv3 = (intvar) v3;
            temp[1] = intv3[0];
            temp[0] = intv3[1];  /* SHOULD BE ANOTHER AUX VARIABLE!!! */

          };
          /* insert the (a = a') for this variable into the globalvar table*/
          insert_globalvar(sptr->name, bdd_identity(bddm, ident), temp[0]);

          /* insert assigned into global_assigned_assoc */
          if (global_assigned_index >= MAX_SYMBOL_SIZE)
               fatal_error("overflow of global variable table");
          if (global_assigned_index != 0)
             bdd_free_assoc(bddm, global_assigned_assoc);
          global_assigned_array[global_assigned_index] = temp[0];
          global_assigned_index++;
          global_assigned_array[global_assigned_index] = 0;
          global_assigned_assoc = 
                   bdd_new_assoc(bddm, global_assigned_array, 0);

          temp[2] = ident;
          ident = bdd_or(bddm, ident, temp[1]);    /* v = v' || nmaintainv */
          v_free_bdd(temp[2]);
          /* &&& -> insert & !gaux into relation */
          temp[2] = ident;
          /* (v = v' || nmaintainv) && !assignedv */
          temp[3] = bdd_not(bddm, temp[0]);
          ident = bdd_and(bddm, ident, temp[3]);
          v_free_bdd(temp[3]);
          v_free_bdd(temp[2]);

          /* update gloval_aux_vars to reflect new auxiliary variable */
          temp[0] = global_nmaintain_vars;
          global_nmaintain_vars =
                      bdd_and(bddm, global_nmaintain_vars, temp[1]);
          v_free_bdd(temp[0]);

          temp[0] = global_identity;
          global_identity = bdd_and(bddm, global_identity, ident);
          v_free_bdd(temp[0]);
        };          
      };
    } else {
      error("Symbol %s being redeclared. Using previous declaration.",
                       symbol_name);
    };
    if (partitionTR) {
      p = p->next;
    } else {
      paux = p;
      p = p->next;
      free(paux);
    };
  };

  /* link local variables to list of variables in the function.
     Used for partitioning transition relations. */
  if (partitionTR) {
    if (inside_function) {
      p = idlist;
      for (paux = p; paux->next != 0; paux = paux->next);
      paux->next = fn_vars;
      fn_vars = p;
    } else {
      p = idlist;
      for (paux = p; paux->next != 0; paux = paux->next);
      paux->next = global_vars;
      global_vars = p;
    };
  };

  /* The transition must be initialized with local identity. */
  if (!is_extern && inside_function) {
    bdd temp;

    temp = R;
    R = bdd_and(bddm, R, local_identity);
    v_free_bdd(temp);
    temp = R;
    R = bdd_and(bddm, R, global_identity);
    v_free_bdd(temp);
  };

  return(VOK);
}


/*
 *
 * selection_statement:
 *   if_condition statement       {reduce_if_stmt();} |
 *   if_condition statement 
 *                ELSE {reduce_else();} statement 
 *                                       {reduce_if_stmt();}
 *   ;
 *
 * if_condition:
 *   IF LP expression RP {reduce_then($3);}
 *   ;
 */

/*
 *   Uses a relation stack to maintain temporary relations. When
 * compiling the then part, the top of the stack contains the original
 * relation && !condition. This relation is used when compiling the else.
 *   At the beginning of the else, the current relation R has the relation
 * after executing the then. It is swapped with the top of the stack, so
 * the else compiles using the original relation.
 *   After the else has been compiled, both relations (the current one and
 * the one on top of the stack) are conjuncted, in accordance with the
 * semantics.
 */
int reduce_then(cond)
typedbdd_ptr cond;
{
  bdd temp[2];

  if (cond->type != VARTYPE_BOOLEAN) {
    fatal_error("non boolean conditional.");
  };

  temp[0] = bdd_not(bddm, cond->tbdd);

  /* push(R & !cond) */
  temp[1] = bdd_and(bddm, R, temp[0]);
  Rpush(temp[1]);

  v_free_bdd(temp[0]);

  /* R = R & cond */
  temp[0] = bdd_and(bddm, R, cond->tbdd);
  v_free_bdd(R);
  R = temp[0];

  v_free_bdd(cond->tbdd);

  return(VOK);
};

int reduce_else()
{
  bdd Raux[2];

  Raux[0] = Rpop();     /* pop(R) */
  Rpush(R);
  R = Raux[0];

  return(VOK);
};

int reduce_if_stmt()
{
  bdd temp[2];

  /* R = R | Rpop */
  temp[0] = R;
  temp[1] = Rpop();

  R = bdd_or(bddm, temp[0], temp[1]);

  v_free_bdd(temp[0]);
  v_free_bdd(temp[1]);

  return(VOK);
};


int saved_position;

extern int srcline;  /* used to avoid counting source lines more than once */

int save_position()
{
  saved_position = filepos;

  /*
   * save_position is called when the token while is read. If > marks
   * the next token to be read, the input file looks like: ... while> ...
   * save_position backtracks until it is: ... >while ...
   */

  /* backtrack until the beginning of while token */
  while (filestart[saved_position] != 'w') saved_position--;
  return(VOK);
};

int reduce_while_condition(cond)
typedbdd_ptr cond;
{
  bdd temp[2];

  if (cond->type != VARTYPE_BOOLEAN) {
    fatal_error("non boolean conditional.");
  };

  /* push(filepos) - save position of beginning of while loop */
  Rpush((bdd) saved_position);
  Rpush((bdd) srcline); /* save line of source code too, since we may come
                           back to this point, srcline should reflect that. */

  /* push(cond) - while loop condition */
  Rpush(bdd_identity(bddm, cond->tbdd));

  /* push(R) - save current relation. Used to check fixpoint convergence */
  Rpush(R);

  temp[0] = bdd_not(bddm, cond->tbdd);
  /* push(R & !cond) - exit condition of while loop.*/
  temp[1] = bdd_and(bddm, R, temp[0]);
  Rpush(temp[1]);

  v_free_bdd(temp[0]);

  /* R = R & cond - update R to reflect we are in the while loop */
  temp[0] = bdd_and(bddm, R, cond->tbdd);
  R = temp[0];
  v_free_bdd(cond->tbdd);
  return(VOK);
};


int reduce_while()
{
  bdd temp[3];

  temp[0] = Rpop(); /* old (R & !cond) */
  temp[1] = bdd_or(bddm, R, temp[0]);
  v_free_bdd(temp[0]);
  v_free_bdd(R);
  R = temp[1];

  /* pop(oldR) */
  temp[0] = Rpop();

  if (R == temp[0]) {
    /* fixpoint found */
    v_free_bdd(temp[0]);

    /* when leaving the loop, make sure R = R & !cond, since we only
       leave the loop when we don't satisfy cond anymore */
    temp[0] = Rpop(); /* cond */
    temp[1] = bdd_not(bddm, temp[0]);
    v_free_bdd(temp[0]);
    temp[0] = bdd_and(bddm, R, temp[1]);
    v_free_bdd(R);
    R = temp[0];

    v_free_bdd(temp[1]);

    temp[0] = Rpop();  /* pop while position from stack - not used */
    temp[0] = Rpop();  /* pop srcline from stack - not used */
  } else {
    /* no fixpoint, iterate again */
    v_free_bdd(temp[0]);
    temp[0] = Rpop();
    /* Restore input position so while loop is parsed again */
    srcline = (long) Rpop();   /* restore srcline too. */
    filepos = (long) Rpop();
  };

  return(VOK);
};


vresult reduce_assignment_internal(); /* forward */

int reduce_time_stmt(t)
int        t;
{
  symbol_ptr   s;
  typedbdd_ptr b;
  int          wc;
  bdd temp[2];
  globalvar_ptr gt;

  /* determine wait count for current wait */
  wc = find_wc(filepos);

  /* insert assignment: waitcount = n */
  b = id_to_bdd(WAITCOUNT_NAME);
  b->tbdd = (bdd) const_to_intvar(wc);
  b->type = VARTYPE_INTEGER;

  reduce_assignment_internal(WAITCOUNT_NAME, b);

  /*
   * generate transition
   */

  /*
   * But first make sure that the global_identity for all variables knows
   * if a new value has been assigned.
   */
  gt = globalvar_table;
  while (gt != 0) {
    /* first, select those transitions in which an assignment was made
       to this variable. */
    temp[1] = bdd_cofactor(bddm, R, gt->assignedv);

    /* now, eliminate all other assignedv variables. We don't care if
       there are assignments for them at this point. */
    bdd_assoc(bddm, global_assigned_assoc);
    temp[0] = bdd_exists(bddm, temp[1]);
    v_free_bdd(temp[1]);

    /* finally, make sure that we don't maintain the value of any
       variable at all, this is done elsewhere. */
    temp[1] = bdd_cofactor(bddm, temp[0], global_nmaintain_vars);
    v_free_bdd(temp[0]);

    temp[0] = gt->global_identity;
    gt->global_identity = bdd_or(bddm, gt->global_identity, temp[1]);
    v_free_bdd(temp[1]);
    v_free_bdd(temp[0]);
    gt = gt->next;
  };

  /*
   * Eliminate all assignedv variables from the current relation. They
   * are not part of the transition.
   */
  if (global_assigned_index != 0) {
    bdd_assoc(bddm, global_assigned_assoc);
    temp[0] = R;
    R = bdd_exists(bddm, R);
    v_free_bdd(temp[0]);
  };

  /* Compute TR = TR or R */

  /*
   * but first, make sure that we don't maintain the value of any variable.
   * this is taken care of later using the global_identity of each var.
   */
  temp[0] = R;
  R = bdd_cofactor(bddm, R, global_nmaintain_vars);
  v_free_bdd(temp[0]);

  temp[1] = TR;
  TR = bdd_or(bddm, R, TR);
  v_free_bdd(temp[1]);

  /* find out the current state waitcount bdd */
  if (lookup_symbol_fullname(WAITCOUNT_NAME, &s) != VOK) {
    if (lookup_symbol(WAITCOUNT_NAME, &s) != VOK) {
      fatal_error("Undeclared wait counter variable %s.", WAITCOUNT_NAME);
    };
  };

  /* Then must reduce the wait statement */

  /* reset R --- R = (wc0 = i) & local_identity*/
  v_free_bdd(R);
  temp[0] = int_to_bdd(wc, id_to_bdd_current(WAITCOUNT_NAME)->tbdd);

  temp[1] = bdd_and(bddm, local_identity, temp[0]);
  v_free_bdd(temp[0]);

  R = bdd_and(bddm, temp[1], global_identity);
  v_free_bdd(temp[1]);

  printf("."); fflush(stdout); /* just to signal that we are alive */
  return(VOK);
}


/*
 * assignment_expression:
 *  identifier ASSIGN expression      {$$ = reduce_assignment($1, $3);}
 *
 * In order to describe all the possible states that can be reached
 * after an assignment, we implement the strongest postcondition formula:
 *
 *            R' = \exists I' ( I = E[I'\I] ^ R[I'\I])
 *
 * R is a global variable.
 */
int reduce_assignment(ident, expr)
char          *ident;
typedbdd_ptr  expr;
{
  if (strstr(ident, WAITCOUNT_NAME))
    fatal_error("Assignments to wait counters are not allowed.");
  reduce_assignment_internal(ident, expr);
  return(VOK);
}

int reduce_assignment_internal(ident, expr)
char          *ident;
typedbdd_ptr  expr;
{
  symbol_ptr   s;
  bdd temp[4];
  bdd assignedv;

  /* checks if ok to assign */

  if (lookup_symbol_fullname(ident, &s) != VOK) {
    if (lookup_symbol(ident, &s) != VOK) {
      fatal_error("Undeclared variable %s.", ident);
    };
  };

  if (s->symtype == SYMTYPE_EXT_VARIABLE) {
    fatal_error("assignment to an extern variable is not allowed.");
  };
  if ((s->symtype != SYMTYPE_LCL_VARIABLE) &&
      (s->symtype != SYMTYPE_GBL_VARIABLE)) {
    fatal_error("assignment to a non variable.");
  };
  if (s->vartype != expr->type) {
    if ((s->vartype != VARTYPE_INTEGER) || (expr->type != VARTYPE_INTLIST)) {
      if ((s->vartype != VARTYPE_BOOLEAN) || (expr->type != VARTYPE_BOOLIST)) {
        fatal_error("assignment between incompatible types.");
      };
    };
  };

  if (s->symtype == SYMTYPE_GBL_VARIABLE) {
    globalvar_ptr gt;

    /*
     * Must update the global identity to reflect the fact that when
     * this process is executing this statement, global identity must
     * not interfere. So, the global identity for this variable cannot
     * apply to the current set of states, represented by R.
     */

    /* determine auxiliary variable for the variable being assigned to */
    if (s->vartype == VARTYPE_BOOLEAN) {
      temp[0] = (bdd) s->value3;         /* nmaintainv variable */
    } else {
      intvar intv;

      if (s->vartype != VARTYPE_INTEGER) 
        fatal_error("unknown type in assignment");
      intv = (intvar) s->value3;         /* nmaintainv variable */
      temp[0] = intv[0];
    };
    /*
     * Make sure that we do maintain the value of the variable before
     * the assignment. This is a requirement of the assignment operation.
     */
    temp[0] = bdd_not(bddm, temp[0]);
    temp[1] = R;
    R = bdd_cofactor(bddm, R, temp[0]);
    v_free_bdd(temp[1]);
    v_free_bdd(temp[0]);

    /*
     * Eliminate the assignedv variable from R. This is needed because
     * the assignedv variable is actually the v_aux used during assigns.
     */
    gt = lookup_globalvar(ident);
    assignedv = gt->assignedv;
    temp[0] = assignedv;
    temp[1] = 0;
    bdd_temp_assoc(bddm, temp, 0);
    bdd_assoc(bddm, -1);
    temp[2] = R;
    R = bdd_exists(bddm, R);
    v_free_bdd(temp[2]);
  };  

  if (s->vartype == VARTYPE_BOOLEAN) {
    /*
     * Boolean variables
     */
    bdd assoc[3];
    bdd I, Iaux;
    int assoc_id;

    I = (bdd) s->value1;
    Iaux = (bdd) s->value3;

    assoc[0] = I;
    assoc[1] = Iaux;
    assoc[2] = 0;
    assoc_id = bdd_new_assoc(bddm, assoc, 1);
    bdd_assoc(bddm, assoc_id);


    if (expr->type == VARTYPE_BOOLEAN) {
      /* A = expr[Iaux\I] */
      temp[0] = bdd_substitute(bddm, expr->tbdd);
      v_free_bdd(expr->tbdd);

      /* B = (I = A) */
      temp[1] = bdd_xnor(bddm, I, temp[0]);
      v_free_bdd(temp[0]);
    } else {
      /* Assign a list of boolean variables using select */
      id_list_ptr s;
      bdd         e;
      bdd         tmp_expr;

      s = (id_list_ptr) expr->tbdd;
      tmp_expr = bdd_zero(bddm);
      while (s != 0) {
        e = (bdd) s->element;

        /* A = expr[i][Iaux\I] */
        temp[0] = bdd_substitute(bddm, e);
        v_free_bdd(e);

        /* B = (I = A) */
        temp[1] = bdd_xnor(bddm, I, temp[0]);
        v_free_bdd(temp[0]);

        /* Disjunct previous select options */
        temp[0] = tmp_expr;
        tmp_expr = bdd_or(bddm, tmp_expr, temp[1]);
        v_free_bdd(temp[0]);
        v_free_bdd(temp[1]);

        s = s->next;
      };
      temp[1] = tmp_expr;
    };
    /* C = r[Iaux\I] */
    temp[0] = bdd_substitute(bddm, R);
    bdd_free_assoc(bddm, assoc_id);
    v_free_bdd(R);

    /* R = \exists Iaux [B & C] */
    assoc[0] = Iaux;
    assoc[1] = 0;
    assoc_id = bdd_new_assoc(bddm, assoc, 0);
    bdd_assoc(bddm, assoc_id);
    R = bdd_rel_prod(bddm, temp[0], temp[1]);
    bdd_free_assoc(bddm, assoc_id);

    v_free_bdd(temp[0]);
    v_free_bdd(temp[1]);
  } else {
    /*
     * Integer variables
     */
    bdd *assoc;
    intvar I, Iaux, Expr;
    int i, n, assoc_id;
    bdd final_expr;

    I    = (intvar) s->value1;
    Iaux = (intvar) s->value3;
    Expr = (intvar) expr->tbdd;

    for(n = 0; I[n] != 0; n++) ; /* determine size of ints */

    /* assoc will substitute Iaux\I */
    assoc = malloc(sizeof(int) * ((2 * n) + 1));
    for(i = 0; i < n; i++) {
      assoc[2 * i] = I[i];
      assoc[(2 * i) + 1] = Iaux[i];
    };
    assoc[2 * i] = 0;

    /*
     * transform a vector of bdds into one bdd: I = expr is actually
     * represented by two bdd vectors for I and expr. We want one bdd
     * expression only. We compute it by using: forall bits in the
     * vector conjunct [I[i] = expr[i][Iaux\I]].
     */
    assoc_id = bdd_new_assoc(bddm, assoc, 1);
    bdd_assoc(bddm, assoc_id);

    if (expr->type == VARTYPE_INTEGER) {
      /* Assign one integer value */
      final_expr = bdd_one(bddm);
      for (i = 0; i < n; i++) {
        /* A = expr[i][Iaux\I] */
        temp[1] = bdd_substitute(bddm, Expr[i]);
        v_free_bdd(Expr[i]);

        /* B = (I[i] = A) */
        temp[0] = bdd_xnor(bddm, I[i], temp[1]);
        v_free_bdd(temp[1]);

        /* final_expr = final_expr & B */
        temp[1] = final_expr;
        final_expr = bdd_and(bddm, temp[0], final_expr);
        v_free_bdd(temp[1]);
        v_free_bdd(temp[0]);
      };
    } else {
      /* Assign a list of integer variables using select */
      id_list_ptr s;
      intvar      e;
      bdd         tmp_expr;

      final_expr = bdd_zero(bddm);
      s = (id_list_ptr) Expr;
      while (s != 0) {
        e = (intvar) s->element;
        tmp_expr = bdd_one(bddm);
        for (i = 0; i < n; i++) {
          /* A = expr[i][Iaux\I] */
          temp[1] = bdd_substitute(bddm, e[i]);
          v_free_bdd(e[i]);

          /* B = (I[i] = A) */
          temp[0] = bdd_xnor(bddm, I[i], temp[1]);
          v_free_bdd(temp[1]);

          /* final_expr = final_expr & B */
          temp[1] = tmp_expr;
          tmp_expr = bdd_and(bddm, temp[0], tmp_expr);
          v_free_bdd(temp[1]);
          v_free_bdd(temp[0]);
        };
        temp[0] = final_expr;
        final_expr = bdd_or(bddm, tmp_expr, final_expr);
        v_free_bdd(temp[0]);
        v_free_bdd(tmp_expr);
        s = s->next;
      };
    };

    /* C = r[Iaux\I] */
    temp[0] = bdd_substitute(bddm, R);
    bdd_free_assoc(bddm, assoc_id);
    v_free_bdd(R);

    /* R = \exists Iaux [final_expr & C] */
    assoc_id = bdd_new_assoc(bddm, Iaux, 0);
    bdd_assoc(bddm, assoc_id);
    R = bdd_rel_prod(bddm, temp[0], final_expr);
    bdd_free_assoc(bddm, assoc_id);
    v_free_bdd(temp[0]);
    v_free_bdd(final_expr);

  };

  if (s->symtype == SYMTYPE_GBL_VARIABLE) {
    /*
     * If this is a global variable, make sure that we flag R as one
     * transition in which a value is assigned to v.
     */
    temp[0] = R;
    R = bdd_and(bddm, R, assignedv);

    v_free_bdd(temp[0]);
  };

  free(expr);
  return(VOK);
};



/*
 * OR_expression:
 *   AND_expression
 *                       {$$ = $1;} |
 *   OR_expression OR AND_expression
 *                       {$$ = reduce_OR_expression($1, $3);}
 */
typedbdd_ptr reduce_OR_expression(expr1, expr2)
typedbdd_ptr expr1, expr2;
{
  typedbdd_ptr result;

  if ((expr1->type != VARTYPE_BOOLEAN) || (expr2->type != VARTYPE_BOOLEAN)) {
    fatal_error("OR on non boolean operators");
  };
  result = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
  result->type = VARTYPE_BOOLEAN;
  result->tbdd = bdd_or(bddm, expr1->tbdd, expr2->tbdd);

  v_free_bdd(expr1->tbdd); free(expr1);
  v_free_bdd(expr2->tbdd); free(expr2);

  return(result);
};


/*
 * conditional_expression:
 *  conditional_expression IMPLIES OR_expression {$$ = reduce_IMPLIES($1,$3);}
 */
typedbdd_ptr reduce_IMPLIES(expr1, expr2)
typedbdd_ptr expr1, expr2;
{
  bdd temp;
  typedbdd_ptr result;

  /* f -> g = !f || g */

  if ((expr1->type != VARTYPE_BOOLEAN) || (expr2->type != VARTYPE_BOOLEAN)) {
    fatal_error("-> on non boolean operators");
  };

  result = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
  result->type = VARTYPE_BOOLEAN;
  temp = bdd_not(bddm, expr1->tbdd);
  v_free_bdd(expr1->tbdd);
  result->tbdd = bdd_or(bddm, temp, expr2->tbdd);

  v_free_bdd(temp);
  free(expr1);
  v_free_bdd(expr2->tbdd); free(expr2);

  return(result);
};


/*
 * conditional_expression:
 *   conditional_expression IFF OR_expression {$$ = reduce_IFF($1, $3);}
 */
typedbdd_ptr reduce_IFF(expr1, expr2)
typedbdd_ptr expr1, expr2;
{
  bdd temp[2];
  typedbdd_ptr r, result, e1, e2;

  /* f <-> g = (f -> g) && (g -> f) */

  if ((expr1->type != VARTYPE_BOOLEAN) || (expr2->type != VARTYPE_BOOLEAN)) {
    fatal_error("<-> on non boolean operators");
  };

  result = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
  result->type = VARTYPE_BOOLEAN;

  e1 = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
  e1->type = VARTYPE_BOOLEAN;
  e1->tbdd = bdd_identity(bddm, expr1->tbdd);
  e2 = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
  e2->type = VARTYPE_BOOLEAN;
  e2->tbdd = bdd_identity(bddm, expr2->tbdd);
  r = reduce_IMPLIES(e1, e2);
  temp[0] = r->tbdd;
  r = reduce_IMPLIES(expr2, expr1);
  result->tbdd = bdd_and(bddm, r->tbdd, temp[0]);
  v_free_bdd(temp[0]);
  v_free_bdd(r->tbdd);
  free(r);

  return(result);
};


/*
 * AND_expression:
 *   equality_expression
 *                       {$$ = $1;} |
 *   AND_expression AND equality_expression
 *                       {$$ = reduce_AND_expression($1, $2);}
 */
typedbdd_ptr reduce_AND_expression(expr1, expr2)
typedbdd_ptr expr1, expr2;
{
  typedbdd_ptr result;

  if ((expr1->type != VARTYPE_BOOLEAN) || (expr2->type != VARTYPE_BOOLEAN)) {
    fatal_error("AND on non boolean operators");
  };
  result = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
  result->type = VARTYPE_BOOLEAN;
  result->tbdd = bdd_and(bddm, expr1->tbdd, expr2->tbdd);
  v_free_bdd(expr1->tbdd); free(expr1);
  v_free_bdd(expr2->tbdd); free(expr2);

  return(result);
};


/*
 * equality_expression:
 *   relational_expression
 *                       {$$ = $1;} |
 *   equality_expression EQUAL  relational_expression
 *                       {$$ = reduce_equality_expression($1, $2, EQUAL);} |
 *   equality_expression NEQUAL relational_expression
 *                       {$$ = reduce_equality_expression($1, $2, NEQUAL);}
 */
typedbdd_ptr reduce_equality_expression(eq_expr, rel_expr, eq_op)
typedbdd_ptr eq_expr, rel_expr;
int          eq_op;
{
  bdd temp[2];

  if (eq_expr->type != rel_expr->type) {
    fatal_error("comparison between different types");
  };

  switch (eq_expr->type) {
    case VARTYPE_BOOLEAN:
      temp[0] = eq_expr->tbdd;
      switch (eq_op) {
        case EQUAL:  eq_expr->tbdd =
                            bdd_xnor(bddm, eq_expr->tbdd, rel_expr->tbdd);
                     break;
        case NEQUAL: eq_expr->tbdd =
                            bdd_xor (bddm, eq_expr->tbdd, rel_expr->tbdd);
                     break;
        default:     fatal_error("Unknown operator for equality expression %d",
                             eq_op);
      };
      v_free_bdd(temp[0]);
      v_free_bdd(rel_expr->tbdd);
      free(rel_expr);
      break;
    case VARTYPE_INTEGER: {
      bdd result;

      switch (eq_op) {
        case EQUAL:
          result = veq(eq_expr->tbdd, rel_expr->tbdd);
          break;
        case NEQUAL:  
          result = vneq(eq_expr->tbdd, rel_expr->tbdd);
          break;
        default:     fatal_error("Unknown operator for equality expression %d",
                             eq_op);
      };
      v_free_int(eq_expr->tbdd);
      v_free_int(rel_expr->tbdd);
      free(rel_expr);
      eq_expr->tbdd = (bdd) result;
      break;
    };
  };

  eq_expr->type = VARTYPE_BOOLEAN;
  return(eq_expr);
};


/*
 *relational_expression:
 *   additive_expression
 *                    {$$ = $1;}  |
 *   relational_expression LESS    additive_expression
 *                    {$$ = reduce_relational_expression($1, $2, LESS);}  |
 *   relational_expression GREATER additive_expression
 *                    {$$ = reduce_relational_expression($1, $2, GREATER);}  |
 *   relational_expression LESSEQ  additive_expression
 *                    {$$ = reduce_relational_expression($1, $2, LESSEQ);}  |
 *   relational_expression GREATEQ additive_expression
 *                    {$$ = reduce_relational_expression($1, $2, GREATEQ);}
 */
typedbdd_ptr reduce_relational_expression(rel_expr, add_expr, rel_op)
typedbdd_ptr rel_expr, add_expr;
int          rel_op;
{
  bdd result;

  if (add_expr->type != VARTYPE_INTEGER) {
    fatal_error("relational operator applied to non integer value");
  };
  if (rel_expr->type != VARTYPE_INTEGER) {
    fatal_error("relational operator applied to non integer value");
  };
  switch (rel_op) {
    case LESS: 
         result = vless(rel_expr->tbdd, add_expr->tbdd);
         break;
    case GREATER:
         result = vgreater(rel_expr->tbdd, add_expr->tbdd);
         break;
    case LESSEQ:
         result = vleq(rel_expr->tbdd, add_expr->tbdd);
         break;
    case GREATEQ:
         result = vgeq(rel_expr->tbdd, add_expr->tbdd);
         break;
    default:
         fatal_error("Unknown operator for relational expression %d",rel_op);
         break;
  };
  v_free_int(rel_expr->tbdd);
  v_free_int(add_expr->tbdd);
  free(add_expr);
  rel_expr->tbdd = result;
  rel_expr->type = VARTYPE_BOOLEAN;
  return(rel_expr);
};


/*
 * additive_expression:
 *   multiplicative_expression
 *                       {$$ = $1;}  |
 *   additive_expression PLUS  multiplicative_expression
 *                       {$$ = reduce_additive_expression($1, $2, PLUS);} |
 *   additive_expression MINUS multiplicative_expression
 *                       {$$ = reduce_additive_expression($1, $2, MINUS);}
 */
typedbdd_ptr reduce_additive_expression(add_expr, mult_expr, add_op)
typedbdd_ptr add_expr, mult_expr;
int          add_op;
{
  bdd temp[2];

  switch (add_op) {
    case PLUS:  if (mult_expr->type != VARTYPE_INTEGER) {
                  fatal_error("addition performed on non integer value");
                };
                if (add_expr->type != VARTYPE_INTEGER) {
                  fatal_error("addition performed on non integer value");
                };
                temp[0] = add_expr->tbdd;
                add_expr->tbdd = (bdd)
                     vadd(add_expr->tbdd, mult_expr->tbdd, bdd_zero(bddm), R);
                v_free_int(temp[0]);
                v_free_int(mult_expr->tbdd);
                free(mult_expr);
                break;
    case MINUS: if (mult_expr->type != VARTYPE_INTEGER) {
                  fatal_error("subtraction performed on non integer value");
                };
                if (add_expr->type != VARTYPE_INTEGER) {
                  fatal_error("subtraction performed on non integer value");
                };
                temp[0] = add_expr->tbdd;
                /* a - m == a + (-m) + 1 */
                add_expr->tbdd = (bdd)
                 vadd(add_expr->tbdd, vneg(mult_expr->tbdd), bdd_one(bddm), R);
                v_free_int(temp[0]);
                v_free_int(mult_expr->tbdd);
                free(mult_expr);
                break;
    default:    fatal_error("Unknown operator for additive expression %d",
                               add_op);
  };
  return(add_expr);
};


/*
 * multiplicative_expression:
 *   unary_expression
 *                    {$$ = $1;} |
 *   multiplicative_expression TIMES unary_expression
 *                    {$$ = reduce_multiplicative_expression($1, $2, TIMES);} |
 *   multiplicative_expression DIV   unary_expression
 *                    {$$ = reduce_multiplicative_expression($1, $2, DIV);} |
 */
typedbdd_ptr reduce_multiplicative_expression(mult_expr, unary_expr, mult_op)
typedbdd_ptr mult_expr, unary_expr;
int          mult_op;
{
  fatal_error("Unknown operator for multiplicative expression %d",mult_op);

  return(mult_expr);
};


/*
 * unary_expression:
 *   primary_expression                {$$ = $1;} |
 *   unary_operator primary_expression {$$ = reduce_unary_expression($1, $2);}
 *
 * unary_operator:
 *   PLUS                  {$$ = PLUS;} |
 *   MINUS %prec UMINUS    {$$ = UMINUS;} |
 *   NOT                   {$$ = NOT;}
 *
 * primary_expression:
 *   identifier            {$$ = id_to_bdd($1);} |
 *   constant              {$$ = const_to_bdd($1);} |
 *   LP expression RP ;    {$$ = $2;}
 */
typedbdd_ptr reduce_unary_expression(unary_op, primary_expr)
int          unary_op;
typedbdd_ptr primary_expr;
{
  bdd temp[2];
  typedbdd_ptr b;

  switch (unary_op) {
    case PLUS:  if (primary_expr->type != VARTYPE_INTEGER) {
                  fatal_error("unary + performed on non integer value");
                };
                break;
    case MINUS: if (primary_expr->type != VARTYPE_INTEGER) {
                  fatal_error("unary - performed on non integer value");
                };
                b = const_to_bdd(0);
                primary_expr->tbdd = (bdd) vadd(vneg(primary_expr->tbdd),
                                       b->tbdd, bdd_one(bddm), R);
                free(b);
                break;
    case NOT:   if (primary_expr->type != VARTYPE_BOOLEAN) {
                  fatal_error("! performed on non boolean value");
                };
                temp[0] = primary_expr->tbdd;
                primary_expr->tbdd = bdd_not(bddm, primary_expr->tbdd);
                v_free_bdd(temp[0]);
                break;
    default:    fatal_error("Unknown operator for unary expression %d",
                               unary_op);
  };

  return(primary_expr);
};


/*
 * expression_list:
 * null                             {$$ = 0;}  |
 * expression                       {$$ = reduce_expression_list(0, $1);} |
 * expression_list COMMA expression {$$ = reduce_expression_list($1, $3);}
 */
typedbdd_ptr reduce_expression_list(oldlist, e)
typedbdd_ptr oldlist, e;
{
  if (e == 0) return(0);

  if (e->type == VARTYPE_BOOLEAN) {
    /*
     * Boolean select: return a list of bdds. Assignment will handle it.
     */
    id_list_ptr exprlist;

    if ((oldlist != 0) && (oldlist->type != VARTYPE_BOOLIST)) {
      fatal_error("Selection between incompatible types.");
    };
    exprlist = (id_list_ptr) malloc(sizeof(struct id_list));
    exprlist->element = (char *) e->tbdd;

    if (oldlist == 0) {
      oldlist = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
      oldlist->type = VARTYPE_BOOLIST;
      oldlist->tbdd = (bdd) exprlist;
      exprlist->next = (id_list_ptr) 0;
    } else {
      exprlist->next = (id_list_ptr) oldlist->tbdd;
      oldlist->tbdd = (bdd) exprlist;
    };
  } else {
    /*
     * Integer select: return a list of intvars. Assignment will handle it.
     */
    id_list_ptr exprlist;

    if ((oldlist != 0) && (oldlist->type != VARTYPE_INTLIST)) {
      fatal_error("Selection between incompatible types.");
    };
    exprlist = (id_list_ptr) malloc(sizeof(struct id_list));
    exprlist->element = (char *) e->tbdd;

    if (oldlist == 0) {
      oldlist = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
      oldlist->type = VARTYPE_INTLIST;
      oldlist->tbdd = (bdd) exprlist;
      exprlist->next = (id_list_ptr) 0;
    } else {
      exprlist->next = (id_list_ptr) oldlist->tbdd;
      oldlist->tbdd = (bdd) exprlist;
    };
  };
  return(oldlist);  
};


/*
 * Looks up a variable name in the symbol table and returns the
 * corresponding bdd (for the next state variable).
 */
typedbdd_ptr id_to_bdd(idname)
char *idname;
{
  symbol_ptr   s;
  typedbdd_ptr b;

  if (lookup_symbol_fullname(idname, &s) != VOK) {
    if (lookup_symbol(idname, &s) != VOK) {
      fatal_error("Undeclared variable %s.", idname);
    };
  };
  b = (typedbdd_ptr) malloc(sizeof(struct typedbdd));

  switch (s->symtype) {
    case SYMTYPE_LCL_VARIABLE:
    case SYMTYPE_GBL_VARIABLE:
      /* Non external variables are denoted by the next state variable */
      switch (s->vartype) {
              /* If we are compiling (statespace == -1), we return the
                 next state bdd. However, if we are parsing a spec
                 (statespace != -1), then every variable is denoted by
                 the current state bdd, because that's how the model
                 checking algorithms want them. */
        case VARTYPE_BOOLEAN:
              if (statespace == -1)
                b->tbdd = bdd_identity(bddm, (bdd) s->value1);
              else
                b->tbdd = bdd_identity(bddm, (bdd) s->value2);
              break;
        case VARTYPE_INTEGER:
              if (statespace == -1)
                b->tbdd = (bdd) v_incref_int((intvar) s->value1);
              else
                b->tbdd = (bdd) v_incref_int((intvar) s->value2);
              break;
        default: fatal_error("variable of unknown type.");
      };
      break;
    case SYMTYPE_ASSERT:
      /* Asserts are denoted by the next state value */
      switch (s->vartype) {
        case VARTYPE_BOOLEAN:
              b->tbdd = bdd_identity(bddm, (bdd) s->value1);
              break;
        default: fatal_error("variable of unknown type.");
      };
      break;
    case SYMTYPE_EXT_VARIABLE:
      /* Extern variables are denoted by the current state variable */
      switch (s->vartype) {
        case VARTYPE_BOOLEAN:
              b->tbdd = bdd_identity(bddm, (bdd) s->value2);
              break;
        case VARTYPE_INTEGER:
              b->tbdd = (bdd) v_incref_int((intvar) s->value2);
              break;
        default: fatal_error("variable of unknown type.");
      };
      break;
    case SYMTYPE_CONSTANT:
              b->tbdd = (bdd) s->value1;
              break;
    default:               fatal_error("translating unknown type to bdd.");
  };
  b->type = s->vartype;
  return(b);
}

/*
 * Looks up a variable name in the symbol table and returns the
 * corresponding bdd (for the current state variable).
 */
typedbdd_ptr id_to_bdd_current(idname)
char *idname;
{
  symbol_ptr   s;
  typedbdd_ptr b;

  if (lookup_symbol_fullname(idname, &s) != VOK) {
    if (lookup_symbol(idname, &s) != VOK) {
      fatal_error("Undeclared variable %s.", idname);
    };
  };
  b = (typedbdd_ptr) malloc(sizeof(struct typedbdd));

  switch (s->symtype) {
    case SYMTYPE_LCL_VARIABLE:
    case SYMTYPE_GBL_VARIABLE:
    case SYMTYPE_EXT_VARIABLE:
      switch (s->vartype) {
        case VARTYPE_BOOLEAN:
              b->tbdd = bdd_identity(bddm, (bdd) s->value2);
              break;
        case VARTYPE_INTEGER:
              b->tbdd = (bdd) v_incref_int((intvar) s->value2);
              break;
        default: fatal_error("variable of unknown type.");
      };
      break;
    default:               fatal_error("translating unknown type to bdd.");
  };
  b->type = s->vartype;
  return(b);
}


/*
 * Returns the bdd that corresponds to the constant.
 *
 * Works only for integer constants. Boolean constants true and false are
 * symbols in Verus.
 */
typedbdd_ptr const_to_bdd(c)
int c;
{
  typedbdd_ptr b;
  b = (typedbdd_ptr) malloc(sizeof(struct typedbdd));
  b->tbdd = (bdd) const_to_intvar(c);
  b->type = VARTYPE_INTEGER;
  return(b);
};


void v_print_state(message, stateset)
char *message;
bdd stateset;
{
  v_print_bdd(message, stateset);
};


/*
 * Given a piece of the source code starting at pos, find the next
 * meaningful character in the program, i.e., from pos on, skip all
 * spaces and comments.
 */
long skip_spaces_comments(pos)
long pos;
{
  long r;
  char c1, c2;

  /* skip spaces from pos */
  for(r = pos; isspace(filestart[r]); r++);

  while (filestart[r] == '/' && filestart[r+1] == '*') {
    /* oops, found a comment, skip it. */

    c1 = c2 = filestart[r++];
    while (!(c1 == '*' && c2 == '/')){
      c1 = c2;
      c2 = filestart[r++];
    }
    for(; isspace(filestart[r]); r++); /* skip spaces after comment */
  };

  if (r > filesize) fatal_error("Syntax error inside spec");

  return(r);
};


void print_spec()
{
  int i, j;

  /*
   * The interval [lastspecpos..filepos] determines *approximately*
   * where the spec being printed is defined in the source code, as
   * shown in the figure below. We must find the exact boundaries.
   */

  /* To do that we must backtrack until we find a ; */
  for (i = filepos; filestart[i] != ';'; i--);

  j = i;
  /* j is where the spec ends, and lastspecpos is where it starts */
  for (i = lastspecpos; i <= j; i++) {
    /* stop at the end of the line or when 55 chars have been printed. */
    if (filestart[i] == '\n' || i - lastspecpos > 55)
      i = j + 1;
    else
      printf("%c", filestart[i]);
  };
  printf("\n");

  /* update lastspecpos */
  lastspecpos = skip_spaces_comments(j+1);
}


/*
 * spec:
 *   SPEC LP ctl_formula RP SEMICOLON {reduce_spec($3);}
 *   ;
 */
int reduce_spec(f)
typedbdd_ptr f;
{
  bdd temp;

  /*
   * A formula is satisfied in the model if it is satisfied in *some*
   * initial state. So we only have to check if the intersection of
   * states satisfying the formula and Init is not false.
   */
  temp = bdd_and(bddm, f->tbdd, Init);
  v_free_bdd(f->tbdd); free(f);
  if (temp == bdd_zero(bddm)) {
    printf("Spec.  is false   : ");
  } else {
    printf("Spec.  is true    : ");

    v_free_bdd(temp);
  };

  print_spec();

  return(VOK);
};


int reduce_min(s, f)
typedbdd_ptr s, f;
{
  int result;

  if ((s->type != VARTYPE_BOOLEAN) || (f->type != VARTYPE_BOOLEAN)) {
    fatal_error("Non boolean expression on minimum computation");
  };

  result = minimum(s->tbdd, f->tbdd);
  free(s);
  free(f);

  if (result == -1) {
    printf("Result is infinity: ");
  } else {
    printf("Result is %7d : ", result);
  };

  print_spec();

  return(VOK);
};


int reduce_max(s, f)
typedbdd_ptr s, f;
{
  int result;

  if ((s->type != VARTYPE_BOOLEAN) || (f->type != VARTYPE_BOOLEAN)) {
    fatal_error("Non boolean expression on maximum computation");
  };

  result = maximum(s->tbdd, f->tbdd);
  free(s);
  free(f);

  if (result == -1) {
    printf("Result is infinity: ");
  } else {
    printf("Result is %7d : ", result);
  };

  print_spec();

  return(VOK);
};


int reduce_mincount(s, c, f)
typedbdd_ptr s, c, f;
{
  int result;

  if ((s->type != VARTYPE_BOOLEAN) ||
      (c->type != VARTYPE_BOOLEAN) ||
      (f->type != VARTYPE_BOOLEAN)) {
    fatal_error("Non boolean expression on mincount computation");
  };

  result = mincount(s->tbdd, c->tbdd, f->tbdd);
  if (result == -1) {
    printf("Result is no path : ");
  } else {
    printf("Result is %7d : ", result);
  };

  print_spec();

  return(VOK);
};


int reduce_maxcount(s, c, f)
typedbdd_ptr s, c, f;
{
  int result;

  if ((s->type != VARTYPE_BOOLEAN) ||
      (c->type != VARTYPE_BOOLEAN) ||
      (f->type != VARTYPE_BOOLEAN)) {
    fatal_error("Non boolean expression on maximum computation");
  };

  result = maxcount(s->tbdd, c->tbdd, f->tbdd);
  if (result == -1) {
    printf("Result is infinity: ");
  } else {
    printf("Result is %7d : ", result);
  };

  print_spec();

  return(VOK);
};


extern void v_print_counterexample();

int reduce_example(s)
typedbdd_ptr s;
{
  if (s->type != VARTYPE_BOOLEAN) {
    fatal_error("Non boolean expression on example computation");
  };

  printf("Example path for  : ");
  print_spec();
  v_print_counterexample(s->tbdd);

  return(VOK);
};





