/*
 * The Verus Model Checking System
 * Copyright (c) 1997 Carnegie Mellon University
 * All Rights Reserved. See file license.txt for details.
 *
 *
 * Grammar for the Verus language
 *
 * Sergio Campos -- 1997
 */

%{
/* declarations */

#include <error.h>
#include <symbols.h>
#include <compile.h>
#include <modcheck.h>
#include <bdd.h>

%}

%union{
  char         *string_t;
  int          int_t;
  id_list_ptr  id_list_t;
  typedbdd_ptr bdd_t;
}

%token            VAR
%token            WAIT
%token            SPEC MIN MAX MINCOUNT MAXCOUNT
%token            IMPLIES IFF AX EX EF AF EG AG E A UNTIL
%token            EXAMPLE
%token            WHILE IF ELSE SELECT
%token            EXTERN BOOLEAN INTEGER
%token            LP RP LB RB LCB RCB
%token            ASSIGN DOTDOT
%token            COLON SEMICOLON COMMA
%token <string_t> IDENTIFIER
%token <int_t>    CONSTANT
%token            OR AND NOT
%token            EQUAL NEQUAL LESS GREATER LESSEQ GREATEQ
%token            PLUS MINUS
%token            UMINUS    /* sets precedence for unary minus */
%token            DOT

/* precedences */

%nonassoc low_prec        /* low_prec is used to avoid conflicts when
                             an expression is part of a temporal expression,
                             e.g. EX a || b. When the parser reads a,
                             EX a_ || b, it is in doubt of reducing EX a,
                             or shifting the ||. low_prec is used to ensure
                             that it shifts the ||.
                           */
%right AX EX EF AF EG AG E A UNTIL EXAMPLE
%left OR IMPLIES IFF
%left AND
%left EQUAL NEQUAL
%left LESS GREATER LESSEQ GREATEQ
%left PLUS MINUS
%left NOT
%left UMINUS    /* sets precedence for unary minus */
%left DOT

/* types */
%type <id_list_t> identifier_list
%type <string_t> identifier
%type <int_t> type_specifier constant unary_operator
%type <bdd_t> primary_expression unary_expression
%type <bdd_t> multiplicative_expression additive_expression
%type <bdd_t> relational_expression equality_expression
%type <bdd_t> AND_expression OR_expression
%type <bdd_t> conditional_expression expression expression_list
%type <bdd_t> temporal_expression

%start translation_unit

%%

translation_unit :
  external_declaration |
  translation_unit external_declaration
  ;

external_declaration:
  function_definition |
  declaration |
  specifications
  ;

function_definition:
  identifier LP identifier_list RP {init_function($1, $3);}
                                   declaration_list compound_statement
    {reduce_function_definition($1, $3, 0, 0);}
  ;

declaration:
  EXTERN type_specifier identifier_list SEMICOLON
    {reduce_declaration($2, $3, 1);} |
         type_specifier identifier_list SEMICOLON
    {reduce_declaration($1, $2, 0);}
  ;

declaration_list:
  null |
  declaration_list declaration
  ;

identifier_list:
  null                             {$$ = 0;}  |
  identifier                       {$$ = reduce_identifier_list(0, $1);} |
  identifier COMMA identifier_list {$$ = reduce_identifier_list($3, $1);}
  ;

type_specifier:
  BOOLEAN {$$ = BOOLEAN;}|
  INTEGER {$$ = INTEGER;}
  ;

compound_statement:
  LCB declaration_list {setup_assocs();} statement_list RCB
  ;

statement_list:
  statement |
  statement_list statement
  ;

statement:
  expression_statement |
  compound_statement |
  selection_statement |
  iteration_statement |
  time_statement
  ;

expression_statement:
  SEMICOLON |
  assignment_expression SEMICOLON
  ;

selection_statement:
  if_condition statement       {reduce_if_stmt();} |
  if_condition statement 
               ELSE {reduce_else();} statement 
                                      {reduce_if_stmt();}
  ;

if_condition:
  IF LP expression RP {reduce_then($3);}
  ;

iteration_statement:
  WHILE                              {save_position();}
        LP expression RP             {reduce_while_condition($4);}
                         statement   {reduce_while();}
  ;

time_statement:
  WAIT LP constant RP       {reduce_time_stmt($3);}
  ;

assignment_expression:
  identifier ASSIGN expression  {reduce_assignment($1, $3);}
  ;

expression:
  conditional_expression   %prec low_prec  {$$ = $1;}
  ;

conditional_expression:
  OR_expression   %prec low_prec           {$$ = $1;} |
  conditional_expression IMPLIES OR_expression {$$ = reduce_IMPLIES($1, $3);} |
  conditional_expression IFF OR_expression {$$ = reduce_IFF($1, $3);}
  ;

OR_expression:
  AND_expression  %prec low_prec
                      {$$ = $1;} |
  OR_expression OR AND_expression
                      {$$ = reduce_OR_expression($1, $3);}
  ;

AND_expression:
  equality_expression   %prec low_prec
                      {$$ = $1;} |
  AND_expression AND equality_expression
                      {$$ = reduce_AND_expression($1, $3);}
  ;

equality_expression:
  relational_expression      %prec low_prec
                      {$$ = $1;} |
  equality_expression EQUAL  relational_expression
                      {$$ = reduce_equality_expression($1, $3, EQUAL);} |
  equality_expression NEQUAL relational_expression
                      {$$ = reduce_equality_expression($1, $3, NEQUAL);}
  ;

relational_expression:
  additive_expression        %prec low_prec
                      {$$ = $1;}  |
  relational_expression LESS    additive_expression
                      {$$ = reduce_relational_expression($1, $3, LESS);}  |
  relational_expression GREATER additive_expression
                      {$$ = reduce_relational_expression($1, $3, GREATER);}  |
  relational_expression LESSEQ  additive_expression
                      {$$ = reduce_relational_expression($1, $3, LESSEQ);}  |
  relational_expression GREATEQ additive_expression
                      {$$ = reduce_relational_expression($1, $3, GREATEQ);}
  ;

additive_expression:
  multiplicative_expression   %prec low_prec
                      {$$ = $1;}  |
  additive_expression PLUS  multiplicative_expression
                      {$$ = reduce_additive_expression($1, $3, PLUS);} |
  additive_expression MINUS multiplicative_expression
                      {$$ = reduce_additive_expression($1, $3, MINUS);}
  ;

multiplicative_expression:
  unary_expression            %prec low_prec
                   {$$ = $1;}
  ;

unary_expression:
  primary_expression                {$$ = $1;} |
  unary_operator primary_expression {$$ = reduce_unary_expression($1, $2);} |
  temporal_expression               {$$ = $1;}
  ;

temporal_expression:
  EX expression                                       {$$ = ex($2);}   |
  AX expression                                       {$$ = ax($2);} |
  EF expression                               {$$ = ef($2, -1, -1);} |
  EF CONSTANT DOTDOT CONSTANT expression      {$$ = ef($5, $2, $4);} |
  AF expression                               {$$ = af($2, -1, -1);} |
  AF CONSTANT DOTDOT CONSTANT expression      {$$ = af($5, $2, $4);} |
  EG expression                               {$$ = eg($2, -1, -1);} |
  EG CONSTANT DOTDOT CONSTANT expression      {$$ = eg($5, $2, $4);} |
  AG expression                               {$$ = ag($2, -1, -1);} |
  AG CONSTANT DOTDOT CONSTANT expression      {$$ = ag($5, $2, $4);} |
  E LB expression UNTIL expression RB {$$ = eu($3, $5, -1, -1);} |
  E LB expression UNTIL CONSTANT DOTDOT CONSTANT expression RB
                                              {$$ = eu($3, $8, $5, $7);} |
  A LB expression UNTIL expression RB {$$ = au($3, $5, -1, -1);} |
  A LB expression UNTIL CONSTANT DOTDOT CONSTANT expression RB
                                              {$$ = au($3, $8, $5, $7);}
  ;

unary_operator:
  PLUS                  {$$ = PLUS;} |
  MINUS %prec UMINUS    {$$ = MINUS;} |
  NOT                   {$$ = NOT;}
  ;

primary_expression:
  identifier                          {$$ = id_to_bdd($1);} |
  constant                            {$$ = const_to_bdd($1);} |
  LP expression RP                    {$$ = $2;} |
  SELECT LCB expression_list RCB      {$$ = $3;}
  ;

expression_list:
  null                             {$$ = 0;}  |
  expression                       {$$ = reduce_expression_list(0, $1);} |
  expression_list COMMA expression {$$ = reduce_expression_list($1, $3);}
  ;

identifier:
  IDENTIFIER {$$ = yylval.string_t;}
  ;

constant:
  CONSTANT   {$$ = yylval.int_t;}
  ;

specifications:
  SPEC {init_modcheck();} spec_list
  ;

spec_list:
  spec |
  spec_list spec
  ;

spec:
  expression SEMICOLON {reduce_spec($1);} |
  MIN      LP expression COMMA expression RP SEMICOLON {reduce_min($3, $5);} |
  MAX      LP expression COMMA expression RP SEMICOLON {reduce_max($3, $5);} |
  MINCOUNT LP expression COMMA expression COMMA expression RP SEMICOLON
                                    {reduce_mincount($3, $5, $7);} |
  MAXCOUNT LP expression COMMA expression COMMA expression RP SEMICOLON
                                    {reduce_maxcount($3, $5, $7);} |
  EXAMPLE expression SEMICOLON               {reduce_example($2);}  ;

null:
  ;

%%

/* programs */

