
/*
 * The Verus Model Checking System
 * Copyright (c) 1997 Carnegie Mellon University
 * All Rights Reserved. See file license.txt for details.
 *
 *
 * The model checker for the Verus tool. Includes model checking and
 * quantitative algorithms. Called from the compiler.
 *
 * Sergio Campos -- 1997
 */

#include <sys/times.h>
#include <time.h>

#include<stdio.h>
#include<strings.h>
#include<stdlib.h>
#include<math.h>
#include<error.h>
#include<compile.h>
#include<symbols.h>
#include<integer.h>
#include<y.tab.h>
#include<modcheck.h>

#include<bdd.h>

extern int verbose, reorder, partitionTR;

extern int lastspecpos;
/*forward*/ extern int skip_spaces_comments();

extern char *filestart;
extern long filesize;
extern long filepos;


/*
 * TR is the transition relation.
 */
bdd TR;

/*
 * Init is the set of initial states.
 */
bdd Init;

/*
 * Reachable is the set of reachable states.
 */
bdd Reachable;

bdd valid_wc;

int    nvars;           /* total number of boolean variables in the program */
int    nwaitvars;       /* total number of boolean variables in wait counter*/
extern int wait_count;  /* total number of wait statements in the program   */
double statespace = -1; /* total number of states in the model.             */
double bdd_statespace = -1; /* total number of states represented by bdds.  */
int gwaitcount = 0;
int gnvars = 0;
int gnwaitvars = 0; /* global number of variables */

double count_bdd();

typedef struct fn_table *fn_table_ptr;
struct fn_table{
  char        *fname;
  int         nwaitvars, nvars, waitcount;
  fn_list_ptr next;     /* position to insert next item */
};


/*forward*/ bdd     step_forward();
/*forward*/ bdd     reachable_from();
/*forward*/ vresult check_deadend();

/*
 * Computes the set of reachable states (if it hasn't been done before).
 */
#define compute_reachable() \
{ \
  Reachable = reachable_from(Init); \
  if (Reachable == bdd_zero(bddm)) { \
    fatal_error("Model has no reachable states!"); \
  }; \
  check_deadend(); \
}

/*
 * The function has been compiled. Prepare to model check.
 */
vresult init_modcheck()
{
  bdd temp[2];
  fn_list_ptr f;
  globalvar_ptr gt;

  fn_list_ptr fn;

  printf("\n\n"); /* prints a \n after all those dots for waits. */

  setup_assocs();
  /*
   * Compose all processes.
   */
  if (functions == 0) fatal_error("No functions defined.");

  f = functions;
  Init = bdd_one(bddm);
  /*
   * The initial value of the transition relation must be global_identity.
   * This bdd describes, for each global variable, the set of states from
   * which that variable is not assigned. Since each transition relation
   * from processes describe when things change, global_identity serves
   * to say when things don't change.
   */
  TR = bdd_one(bddm);
  gt = globalvar_table;
  while (gt != 0) {
    /* TR = TR & var.global_identity */
    temp[0] = TR;
    TR = bdd_and(bddm, TR, gt->global_identity);
    v_free_bdd(temp[0]);

    v_free_bdd(gt->global_identity);

    gt = gt->next;
  };

  valid_wc = bdd_one(bddm);
  while (f != 0) {
    if (!partitionTR) {
      temp[0] = TR;
      TR = bdd_and(bddm, f->TR, TR);
      v_free_bdd(temp[0]);
      v_free_bdd(f->TR);
    };
    temp[0] = Init;
    Init = bdd_and(bddm, f->Init, Init);
    v_free_bdd(temp[0]);
    v_free_bdd(f->Init);
    gwaitcount += f->waitcount;
    gnvars     += f->nvars;
    gnwaitvars += f->nwaitvars;

    /* Determine which values of the wait counters are valid */
    {
      char name[100];
      typedbdd_ptr t1, t2;

      strcpy(name, f->fname);
      strcat(name, ".");
      strcat(name, WAITCOUNT_NAME);
      t1 = id_to_bdd_current(name);
      /* t2 is the relation for <f->name>._wc <= <f->waitcount> */
      t2 = 
         reduce_relational_expression(t1, const_to_bdd(f->waitcount), LESSEQ);
      valid_wc = bdd_and(bddm, valid_wc, t2->tbdd);
      v_free_bdd(t2->tbdd); free(t2);
    };

    f = f->next;
    if (reorder) {
      bdd_dynamic_reordering(bddm, bdd_reorder_sift);
      printf("Size of transition relation before reorder: %.0f.",
                           (double) bdd_size(bddm, TR, 1));
      fflush(stdout);
      bdd_reorder(bddm);
      printf(" After: %.0f bdd nodes.\n",
                           (double) bdd_size(bddm, TR, 1));
    };
  };

  if (partitionTR) {
    /*
     * insert the global_identity v=v' as another process. This fake
     * process determines when global variables do *not* change value.
     */
    fn = (fn_list_ptr) malloc(sizeof(struct fn_list));
    fn->fname     = "v=v'";
    fn->TR        = TR;
    fn->Init      = bdd_one(bddm);
    fn->nwaitvars = 0;
    fn->nvars     = 0;
    fn->waitcount = 0;
    fn->next      = functions;
    functions     = fn;
  };

  if (reorder) {
    double oldsize;

    printf("Reordering bdd variables...\n");
    bdd_dynamic_reordering(bddm, bdd_reorder_sift);
    /* bdd_dynamic_reordering(bddm, bdd_reorder_stable_window3);*/
    oldsize = 0;
    while (oldsize != bdd_size(bddm, TR, 1)) {
      oldsize = bdd_size(bddm, TR, 1);
      printf("Size of transition relation is %.0f bdd nodes.\n", oldsize);
      bdd_reorder(bddm);
    };

    /* bdd_reorder has to be explicitly turned off or 
       it will happen at random */
    bdd_dynamic_reordering(bddm, bdd_reorder_none);
  };

{
  struct tms buffer;

  times(&buffer);
  printf("Time to construct the model: User : %8g s",
	 buffer.tms_utime/(double)CLK_TCK);
  printf("  System   : %8g s\n\n",
	 buffer.tms_stime/(double)CLK_TCK);
}

  if (statespace == -1) {

    /* Compute state space. wait_count-1 is the number of waits */
    statespace = gwaitcount * pow(2.0, (double) gnvars);
    /*
     * bdd_satisfying_fractions thinks that all possible values of the
     * wait counter are valid. So, to compute the correct number of
     * states in b we must use a slightly larger state space.
     */
    bdd_statespace = pow(2.0, (double) (gnwaitvars + gnvars));
  };

  temp[0] = Init;
  Init = step_forward(Init);
  v_free_bdd(temp[0]);

  if (verbose)
      printf("Initial state set: states %.0f, bdd size %.0f.\n",
             count_bdd(Init), (double) bdd_size(bddm, Init, 1));
  if (verbose == 2) v_print_state("Init", Init);

  compute_reachable();

  /* Set up pointer in the input file to print specifications. */
  lastspecpos = skip_spaces_comments(filepos);

  return(VOK);
};


double count_bdd(b)
bdd b;
{
  return(bdd_statespace * bdd_satisfying_fraction(bddm, b));
};


/*
 * step_forward: Given a set of states f, we find the successors of f by:
 *
 * - find the set of transitions from f, i.e., the transitions in which
 *   the current state is f. This is Z1 = f & TR.
 * - Existentially quantify out the current state in those transitions.
 *   This is Z2 = \exists (current_state) Z1.
 *   Note: These two steps are performed by a relational product, which
 *         is much more efficient than doing it separately.
 * - Z2 contains the sets of states that are successors of f, but the
 *   bdd variables in Z2 are the next state vars. Substitute them for
 *   the current state vars using bdd_substitute.
 */
bdd step_forward(f)
bdd f;
{
  bdd temp[3];
  fn_list_ptr fn;

  if (!partitionTR) {
    bdd_assoc(bddm, current_assoc);      /* quantify current state     */
    temp[0] = bdd_rel_prod(bddm, f, TR); /* out of (f & TR)            */

    bdd_assoc(bddm, next_assoc);         /* shift vars back to current */
    temp[1] = bdd_substitute(bddm, temp[0]);
    v_free_bdd(temp[0]);
  } else {
    temp[2] = bdd_one(bddm);
    fn = functions;
    while (fn != 0) {
      temp[0] = bdd_reduce(bddm, fn->TR, f);
      temp[1] = temp[2];
      temp[2] = bdd_and(bddm, temp[2], temp[0]);
      v_free_bdd(temp[0]);
      v_free_bdd(temp[1]);
      if (verbose) {
        printf("   step_forward: function %s, states %.0f, bdd size %.0f.\n",
          fn->fname, count_bdd(temp[2]), (double) bdd_size(bddm, temp[2], 1));
      };
      fn = fn->next;
    };
    bdd_assoc(bddm, current_assoc);
    temp[1] = temp[2];
    temp[2] = bdd_rel_prod(bddm, f, temp[2]);
    v_free_bdd(temp[1]);
    bdd_assoc(bddm, next_assoc);         /* shift vars back to current */
    temp[1] = bdd_substitute(bddm, temp[2]);
    v_free_bdd(temp[2]);
  };
  return(temp[1]);
};


/*
 * step_backwards: Given a set of states f, we find the predecessors of f:
 *
 * - Translate the variables in f to the next state using bdd_substitute (Z1).
 * - find the set of transitions ending in f, i.e., the transitions in
 *   which the next state is f. This is Z2 = Z1 & TR.
 * - Existentially quantify out the next state in those transitions.
 *   This is Z3 = \exists (next_state) Z1.
 *   Note: These two steps are performed by a relational product, which
 *         is much more efficient than doing it separately.
 *
 * Note: Going backwards is usually more expensive than going forwards.
 * Because of this, we use the context bdd to simplify the set of states
 * and therefore speed up the computation. Context is normally the current
 * relation, that is, transitions that do not belong to the current
 * relation do not affect the result anyway, and are "reduced" away from
 * the operation.
 */
bdd step_backwards(f, context)
bdd f, context;
{
  bdd temp[3], fcurrent;
  fn_list_ptr fn;

  if (!partitionTR) {
    bdd_assoc(bddm, current_assoc);      /* replace current state vars */
    temp[0] = bdd_substitute(bddm, f);   /* with next state vars       */

    /* Keep only transitions into states satisfying the context */
    temp[2] = bdd_reduce(bddm, TR, context);

    bdd_assoc(bddm, next_assoc);               /* find which transitions  */
    temp[1] = bdd_rel_prod(bddm, temp[0], temp[2]);
                                               /* satisfy this next state */
    v_free_bdd(temp[0]);
    v_free_bdd(temp[2]);
    return(temp[1]);
  } else {
    bdd_assoc(bddm, current_assoc);      /* replace current state vars */
    fcurrent = bdd_substitute(bddm, f);   /* with next state vars       */
    temp[2] = bdd_one(bddm);
    fn = functions;
    while (fn != 0) {
      temp[0] = bdd_reduce(bddm, fn->TR, fcurrent);
      temp[1] = temp[0];
      temp[0] = bdd_reduce(bddm, temp[0], context);
      v_free_bdd(temp[1]);
      temp[1] = temp[2];
      temp[2] = bdd_and(bddm, temp[0], temp[2]);
      v_free_bdd(temp[0]);
      v_free_bdd(temp[1]);
      if (verbose == 2) {
        printf("   step_backwards: function %s, bdd size %.0f.\n",
           fn->fname, (double) bdd_size(bddm, temp[0], 1));
      };
      fn = fn->next;
    };
    bdd_assoc(bddm, next_assoc);
    temp[0] = bdd_rel_prod(bddm, fcurrent, temp[2]);
    v_free_bdd(temp[2]);
    return(temp[0]);
  };
};


/*
 * reachable_from returns the set of states that are reachable from f.
 */
bdd reachable_from(f)
bdd f;
{
  bdd R, Rp, frontier;
  bdd temp[2];
  int i = 0;

  R  = bdd_zero(bddm);
  Rp = bdd_identity(bddm, f);
  frontier = bdd_identity(bddm, f);
  while (R != Rp) {
    v_free_bdd(R);
    R = Rp;
    if (verbose) {
      printf("Reachable: iteration %d, states %.0f, bdd size %.0f.\n",
             i++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
      if (verbose == 2) v_print_state("", frontier);
    };
    temp[0] = frontier;
    frontier = step_forward(frontier);
    v_free_bdd(temp[0]);
    temp[0] = Rp;
    Rp = bdd_or(bddm, Rp, frontier);
  };

  return(Rp);
};


/*
 * bdd list is used to keep track of states reached backwards from the
 * end of the counterexample. It is needed in order to print the
 * counterexample from beginning to end, instead of starting at the end.
 */
typedef struct bddlist *bddlist_ptr;
struct bddlist{
  bdd         b;
  bddlist_ptr next;
};


vresult v_print_counterexample(s)
bdd s;
{
  bdd temp[3];
  bddlist_ptr bl, counterexample;
  int i;

  s = bdd_and(bddm, s, Reachable);
  if (s == bdd_zero(bddm)) {
    printf("State set is not reachable\n");
    return(VOK);
  };

  counterexample = 0;
  temp[0] = bdd_and(bddm, s, Init);
  while (temp[0] == bdd_zero(bddm)) {         /* until we reach start do */
    bl = (bddlist_ptr) malloc(sizeof(struct bddlist));

    /* set up current_assoc to ensure that only current state will be shown */
/*    bdd_assoc(bddm, current_assoc);*/
/*    bl->b = bdd_satisfy_support(bddm, s);*/     /* find one state in s and
                                                 put it into the list */
    bl->b = s;
    bl->next = counterexample;
    counterexample = bl;
    s = step_backwards(bl->b, bdd_one(bddm)); /* and go back one step */
    temp[0] = s;
    s = bdd_and(bddm, s, Reachable);
    v_free_bdd(temp[0]);
    temp[0] = bdd_and(bddm, s, Init);
  };
  bl = (bddlist_ptr) malloc(sizeof(struct bddlist));
  /* set up current_assoc to ensure that only current state will be shown */
  bdd_assoc(bddm, current_assoc);
  bl->b = bdd_satisfy_support(bddm, temp[0]);
  bl->next = counterexample;
  counterexample = bl;
  v_free_bdd(temp[0]);

  /*
   * At this point counterexample is a list of bdds such that:
   *   1. each bdd is only one state;
   *   2. It starts in Init;
   *   3. It leads to the original s.
   *
   *  Now print this list
   */
   i = 1;
   temp[1] = bdd_one(bddm);
   while (counterexample != 0) {
     printf("state %3d:", i); i++;
     temp[0] = bdd_and(bddm, counterexample->b, temp[1]);
     v_free_bdd(counterexample->b);
     v_free_bdd(temp[1]);
     bdd_assoc(bddm, current_assoc);
     temp[2] = bdd_satisfy_support(bddm, temp[0]);
     v_print_state("", temp[2]);
     temp[1] = step_forward(temp[2]);
     v_free_bdd(temp[2]);
     bl = counterexample;
     counterexample = counterexample->next;
     free(bl);
   };
   return(VOK);     
}


/*
 * check_deadend() checks to see if there are reachable states that do
 * not have outgoing transitions.
 */
vresult check_deadend()
{
  bdd temp[2];

  /* find those states that have a successor */
  temp[0] = step_backwards(bdd_one(bddm), bdd_one(bddm));

  /* Now find those that do not... */
  temp[1] = bdd_not(bddm, temp[0]);
  v_free_bdd(temp[0]);

  /* And that are reachable */
  temp[0] = bdd_and(bddm, Reachable, temp[1]);
  v_free_bdd(temp[1]);

  /* if there is any, print one of them. */
  if (temp[0] != bdd_zero(bddm)) {
    typedbdd_ptr tt, r;

    printf("Reachable state without outgoing transition.\n");
    v_print_counterexample(temp[0]);
    v_free_bdd(temp[0]);

    /*
     * We must restrict the set of reachable states to those that
     * don't get stuck in a deadend state.
     */
    tt = malloc(sizeof(struct typedbdd));
    tt->type = VARTYPE_BOOLEAN;
    tt->tbdd = bdd_one(bddm);
    r = eg(tt,-1,-1);    /* find those states that lead to infinite paths. */

    temp[1] = Reachable;
    Reachable = bdd_and(bddm, Reachable, r->tbdd);
    v_free_bdd(temp[1]);
    temp[1] = Init;
    Init = bdd_and(bddm, Init, r->tbdd);
    v_free_bdd(temp[1]);
    v_free_bdd(r->tbdd);
    free(r);
  };

  return(VOK);
};


/*
 * Computes the minimum delay from a state in start to any state in final.
 */
int minimum(start, final)
bdd start, final;
{
  int i, iteration;
  bdd R, Rp, frontier;
  bdd temp[2];

  iteration = 0;
  i = 0;
  R = bdd_and(bddm, start, Reachable);
  v_free_bdd(start);
  frontier = step_forward(R);
  Rp = bdd_or(bddm, R, frontier);
  temp[0] = bdd_and(bddm, R, final);
  if (verbose) {
    printf("Minimum: iteration %d, states %.0f, bdd size %.0f.\n",
           iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
    if (verbose == 2) v_print_state("", Rp);
  };
  while ((R != Rp) && (temp[0] == bdd_zero(bddm))) {
    i++;
    v_free_bdd(R);
    R = Rp;
    temp[0] = frontier;
    frontier = step_forward(frontier);
    v_free_bdd(temp[0]);
    Rp = bdd_or(bddm, frontier, Rp);
    if (verbose) {
      printf("Minimum: iteration %d, states %.0f, bdd size %.0f.\n",
             iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
      if (verbose == 2) v_print_state("", Rp);
    };
    temp[0] = bdd_and(bddm, R, final);
  };
  v_free_bdd(R);
  v_free_bdd(Rp);
  if (temp[0] != bdd_zero(bddm)) {
    v_free_bdd(temp[0]);
    return(i);
  } else {
    return(-1); /* infinity */
  };
};


/*
 * Computes the maximum delay from a state in f to any state in g.
 */
int maximum(start, final)
bdd start, final;
{
  int i, iteration;
  bdd R, Rp, not_final;
  bdd temp[2];

  iteration = 0;
  temp[0] = bdd_not(bddm, final);
  not_final = bdd_and(bddm, temp[0], Reachable);
  v_free_bdd(temp[0]);
  v_free_bdd(final);
  i = 0;

  R = bdd_one(bddm);
  Rp = bdd_identity(bddm, not_final);
  temp[0] = bdd_and(bddm, Rp, start);
  if (verbose) {
    printf("Maximum: iteration %d, states %.0f, bdd size %.0f.\n",
           iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
    if (verbose == 2) v_print_state("", Rp);
  };
  while ((R != Rp) && (temp[0] != bdd_zero(bddm))) {
    v_free_bdd(temp[0]);
    i++;
    v_free_bdd(R);
    R = Rp;
    temp[0] = step_backwards(Rp, not_final);
    Rp = bdd_and(bddm, temp[0], not_final);
    v_free_bdd(temp[0]);
    if (verbose) {
      printf("Maximum: iteration %d, states %.0f, bdd size %.0f.\n",
             iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
      if (verbose == 2) v_print_state("", Rp);
    };
    temp[0] = bdd_and(bddm, Rp, start);
  };
  v_free_bdd(temp[0]);
  v_free_bdd(start);
  v_free_bdd(not_final);
  if (R == Rp) {
    v_free_bdd(R);
    v_free_bdd(Rp);
    return(-1); /* infinity */
  } else {
    v_free_bdd(R);
    v_free_bdd(Rp);
    return(i);
  };
};


/*
 * Computes the minimum delay from a state in start to any state in final.
 */
int mincount(start, condition, final)
bdd start, condition, final;
{
  int i, iteration;
  bdd R, Rp;
  bdd start_cond, start_ncond;
  bdd temp[3];

  start_cond = bdd_and(bddm, start, condition);
  temp[0] = bdd_not(bddm, condition);
  start_ncond = bdd_and(bddm, start, temp[0]);
  v_free_bdd(temp[0]);
  v_free_bdd(start);
  temp[0] = start_cond;
  start_cond = bdd_and(bddm, start_cond, Reachable);
  v_free_bdd(temp[0]);
  temp[0] = start_ncond;
  start_ncond = bdd_and(bddm, start_ncond, Reachable);
  v_free_bdd(temp[0]);

  iteration = 0;
  i = 0;
  R = bdd_zero(bddm);
  Rp = bdd_identity(bddm, start_ncond);
  if (verbose) {
     printf("Mincount: iteration %d, states %.0f, bdd size %.0f.\n",
         iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
     if (verbose == 2) v_print_bdd("", Rp);
  };
  do {
    do {
      temp[0] = bdd_and(bddm, Rp, final);
      if (temp[0] != bdd_zero(bddm)) {
        v_free_bdd(temp[0]);
        v_free_bdd(start_cond);
        v_free_bdd(start_ncond);
        v_free_bdd(condition);
        v_free_bdd(final);
        return(i);
      };
      v_free_bdd(temp[0]);
      v_free_bdd(R);
      R = Rp;
      temp[0] = step_forward(Rp);
      temp[1] = bdd_not(bddm, condition);
      temp[2] = bdd_and(bddm, temp[0], temp[1]);
      v_free_bdd(temp[0]);
      v_free_bdd(temp[1]);
      Rp = bdd_or(bddm, temp[2], Rp);
      v_free_bdd(temp[2]);
    } while (R != Rp);
    temp[0] = step_forward(Rp);
    temp[1] = bdd_and(bddm, temp[0], condition);
    v_free_bdd(temp[0]);
    temp[0] = Rp;
    Rp = bdd_or(bddm, temp[1], Rp);
    if (i == 0) {
      temp[0] = Rp;
      Rp = bdd_or(bddm, Rp, start_cond);
      v_free_bdd(temp[0]);
    };
    if (verbose) {
       printf("Mincount: iteration %d, states %.0f, bdd size %.0f.\n",
           iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
       if (verbose == 2) v_print_state("", Rp);
    };
    i++;
  } while (R != Rp);
  v_free_bdd(start_cond);
  v_free_bdd(start_ncond);
  v_free_bdd(condition);
  v_free_bdd(final);
  return(-1); /* infinity */
};


/*
 * Computes the maximum delay from a state in f to any state in g.
 */
int maxcount(start, condition, final)
bdd start, condition, final;
{
  int i, iteration;
  bdd R, R1, Rp, not_final, not_condition;
  bdd temp[3];

  iteration = 0;
  i = 1;
  temp[0] = condition;
  not_condition = bdd_not(bddm, condition);
  condition = bdd_and(bddm, condition, Reachable);
  v_free_bdd(temp[0]);
  temp[0] = not_condition;
  not_condition = bdd_and(bddm, not_condition, Reachable);
  v_free_bdd(temp[0]);
  not_final = bdd_not(bddm, final);
  v_free_bdd(final);
  R1 = bdd_zero(bddm);
  Rp = bdd_identity(bddm, condition);
  if (verbose) {
     printf("Maxcount: iteration %d, states %.0f, bdd size %.0f.\n",
         iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
     if (verbose == 2) v_print_state("", Rp);
  };
  do {
    v_free_bdd(R1);
    R1 = Rp;
    do {
      R = Rp;
      temp[0] = bdd_and(bddm, not_final, not_condition);
      temp[1] = step_backwards(Rp, temp[0]);
      v_free_bdd(temp[0]);
      temp[0] = bdd_and(bddm, temp[1], not_final);
      v_free_bdd(temp[1]);
      temp[1] = bdd_and(bddm, temp[0], not_condition);
      v_free_bdd(temp[0]);
      Rp = bdd_or(bddm, temp[1], Rp);
      v_free_bdd(temp[1]);
    } while (R != Rp);
    temp[0] = bdd_and(bddm, Rp, start);
    if (temp[0] == bdd_zero(bddm)) {
      v_free_bdd(temp[0]);
      v_free_bdd(not_condition);
      v_free_bdd(condition);
      v_free_bdd(not_final);
      v_free_bdd(start);
      return(i - 1);
    };
    v_free_bdd(temp[0]);
 
    temp[0] = step_backwards(Rp, not_final);
    Rp = bdd_and(bddm, temp[0], condition);
    v_free_bdd(temp[0]);
    temp[0] = Rp;
    Rp = bdd_and(bddm, Rp, not_final);
    if (verbose) {
       printf("Maxcount: iteration %d, states %.0f, bdd size %.0f.\n",
           iteration++, count_bdd(Rp), (double) bdd_size(bddm, Rp, 1));
       if (verbose == 2) v_print_state("", Rp);
    };
    i++;
  } while (R1 != Rp);
  v_free_bdd(not_condition);
  v_free_bdd(condition);
  v_free_bdd(not_final);
  v_free_bdd(start);
  return(-1); /* infinity */
};


/*
 * CTL_expression:
 *   EX CTL_expression {$$ = ex($2);}
 *   ;
 */
typedbdd_ptr ex(f)
typedbdd_ptr f;
{
  bdd temp;

  if (statespace == -1) fatal_error("CTL formula outside specification");
  if (f->type != VARTYPE_BOOLEAN) {
    fatal_error("Non boolean expression on CTL formula: EX");
  };

  /* The set of predecessors of f is the set of states that have a
  successor in f. */

  temp = step_backwards(f->tbdd, bdd_one(bddm));
  v_free_bdd(f->tbdd);
  f->tbdd = bdd_and(bddm, temp, Reachable);
  v_free_bdd(temp);
  return(f);
};


/*
 * CTL_expression:
 *   AX CTL_expression {$$ = ax($2);}
 *   ;
 */
typedbdd_ptr ax(f)
typedbdd_ptr f;
{
  bdd temp[2];

  if (statespace == -1) fatal_error("CTL formula outside specification");
  if (f->type != VARTYPE_BOOLEAN) {
    fatal_error("Non boolean expression on CTL formula: AX");
  };

  /* AX f = !EX !f */

  temp[0] = f->tbdd;
  f->tbdd = bdd_not(bddm, f->tbdd);
  v_free_bdd(temp[0]);
  temp[1] = f->tbdd;
  f = ex(f);
  temp[1] = f->tbdd;
  temp[0] = bdd_not(bddm, f->tbdd);
  v_free_bdd(temp[1]);
  f->tbdd = bdd_and(bddm, temp[0], Reachable);
  v_free_bdd(temp[0]);
  return(f);
};


/*
 * CTL_expression:
 *   EF CTL_expression {$$ = ef($2);}
 *   ;
 */
typedbdd_ptr ef(f, min, max)
int min, max;
typedbdd_ptr f;
{
  typedbdd_ptr tt;

  if (f->type != VARTYPE_BOOLEAN) {
    fatal_error("Non boolean expression on CTL formula: AG");
  };

  /* EF f = E [ true U f ] */

  tt = malloc(sizeof(struct typedbdd));
  tt->type = VARTYPE_BOOLEAN;
  tt->tbdd = bdd_one(bddm);

  if (verbose) printf("Computing EF...\n");

  return(eu(tt, f, min, max));
};


/*
 * CTL_expression:
 *   AF CTL_expression {$$ = af($2);}
 *   ;
 */
typedbdd_ptr af(f, min, max)
int min, max;
typedbdd_ptr f;
{
  bdd temp[2];

  if (f->type != VARTYPE_BOOLEAN) {
    fatal_error("Non boolean expression on CTL formula: AG");
  };

  /* AF f = !EG !f */

  temp[0] = f->tbdd;
  f->tbdd = bdd_not(bddm, f->tbdd);
  v_free_bdd(temp[0]);

  f = eg(f, min, max);

  temp[0] = f->tbdd;
  f->tbdd = bdd_not(bddm, f->tbdd);
  f->tbdd = bdd_and(bddm, f->tbdd, Reachable);
  v_free_bdd(temp[0]);

  return(f);
};


/*
 * CTL_expression:
 *   EG CTL_expression {$$ = eg($2);}
 *   ;
 */
typedbdd_ptr eg(f, min, max)
int min, max;
typedbdd_ptr f;
{

  bdd temp[2];
  bdd oldz;
  typedbdd_ptr z;
  int iteration = 0;
  int i;

  if (statespace == -1) fatal_error("CTL formula outside specification");
  if (f->type != VARTYPE_BOOLEAN) {
    fatal_error("Non boolean expression on CTL formula: until");
  };
  if ((max != -1) && (min > max || min < 0))
      fatal_error("Invalid bounds on RTCTL formula");

  /* EG f = f && EX EG f*/

  oldz = bdd_zero(bddm);
  /* The states that satisfy EG f are those that satisfy f and... */
  z = f;          /* recycle space used by f. */
  temp[0] = z->tbdd;
  z->tbdd = bdd_and(bddm, z->tbdd, Reachable);
  v_free_bdd(temp[0]);

  if (verbose) {
    printf("EG: iteration %d, states on frontier %.0f, bdd size %.0f.\n",
           iteration++, count_bdd(z->tbdd),
           (double) bdd_size(bddm, z->tbdd, 1));
    if (verbose == 2) v_print_state("", z->tbdd);
  };
  if (max != -1) i = max;
            else i = min + 1;
  /*
   * - (oldz != z->tbdd) checks for fixpoint,
   * - (i > min) checks for RTCTL bounds (notice that in the
   *   unbounded case i is set to min+1 and it is not decremented).
   */
  while ((oldz != z->tbdd) && (i > min)) {
    v_free_bdd(oldz);
    oldz = bdd_identity(bddm, z->tbdd);
    /* ... also have a successor that satisfies f. */
    z = ex(z);
    temp[0] = z->tbdd;
    z->tbdd = bdd_and(bddm, oldz, z->tbdd);
    v_free_bdd(temp[0]);
    if (verbose) {
      printf("EG: iteration %d, states on frontier %.0f, bdd size %.0f.\n",
             iteration++, count_bdd(z->tbdd),
             (double) bdd_size(bddm, z->tbdd, 1));
      if (verbose == 2) v_print_state("", z->tbdd);
    };
    if (max != -1) i--;
  };

  if (max != -1) {
    /*
     * Once we've checked inside the bounds step min times to check
     * outside the bounds.
     */
    oldz = bdd_zero(bddm);
    i = min;
    while ((oldz != z->tbdd) && (i > 0)) {
      v_free_bdd(oldz);
      oldz = bdd_identity(bddm, z->tbdd);
      /* ... also have a successor that satisfies f. */
      z = ex(z);
      if (verbose) {
        printf("EG: Outside bounds: iteration %d, states on frontier %.0f, bdd size %.0f.\n",
	       iteration++, count_bdd(z->tbdd),
	       (double) bdd_size(bddm, z->tbdd, 1));
	if (verbose == 2) v_print_state("", z->tbdd);
      };
      if (max != -1) i--;
    };
  };

  return(z);
};


/*
 * CTL_expression:
 *   AG CTL_expression {$$ = ag($2);}
 *   ;
 */
typedbdd_ptr ag(f, min, max)
int min, max;
typedbdd_ptr f;
{
  bdd temp[2];

  if (statespace == -1) fatal_error("CTL formula outside specification");
  if (f->type != VARTYPE_BOOLEAN) {
    fatal_error("Non boolean expression on CTL formula: AG");
  };

  /* AG f = !EF !f */

  temp[0] = f->tbdd;
  f->tbdd = bdd_not(bddm, f->tbdd);
  v_free_bdd(temp[0]);

  f = ef(f, min, max);

  temp[0] = f->tbdd;
  f->tbdd = bdd_not(bddm, f->tbdd);
  v_free_bdd(temp[0]);

  return(f);
};


/*
 * CTL_expression:
 *   E [ CTL_expression U CTL_expression ] {$$ = eu($3, $5);}
 *   ;
 */
typedbdd_ptr eu(f, g, min, max)
int min, max;
typedbdd_ptr f, g;
{
  bdd temp[2];
  bdd oldz, z;
  typedbdd_ptr frontier;
  int iteration = 0;
  int i;

  if (statespace == -1) fatal_error("CTL formula outside specification");
  if ((f->type != VARTYPE_BOOLEAN) || (g->type != VARTYPE_BOOLEAN)) {
    fatal_error("Non boolean expression on CTL formula: until");
  };
  if ((max != -1) && (min > max || min < 0))
      fatal_error("Invalid bounds on RTCTL formula");

  /* E[f U g] = fix z. [ g || (f && EX z)]  */

  oldz = bdd_zero(bddm);
  /* The states that satisfy E[f U g] are those that satisfy g or... */
  z = g->tbdd;
  z = bdd_and(bddm, z, Reachable);
  v_free_bdd(g->tbdd);
  frontier = g;   /* recycle space used by g */
  frontier->tbdd = bdd_identity(bddm, z);
  if (verbose) {
    printf("EU: iteration %d, states on frontier %.0f, bdd size %.0f.\n",
           iteration++, count_bdd(frontier->tbdd),
           (double) bdd_size(bddm, frontier->tbdd, 1));
    if (verbose == 2) v_print_state("", frontier->tbdd);
  };
  if (max != -1) i = max;
            else i = min + 1;
  /*
   * - (frontier->tbdd != bdd_zero(bddm)) checks for fixpoint,
   * - (i > min) checks for RTCTL bounds (notice that in the
   *   unbounded case i is set to min+1 and it is not decremented).
   */
  while ((frontier->tbdd != bdd_zero(bddm)) && (i > min)) {
    v_free_bdd(oldz);
    oldz = z;
    /* ...those that satisfy f and have a transition into a state that
       satisfies E[f U g]. */
    /* z = z || (f->tbdd && ex(frontier)) */
    frontier = ex(frontier);
    temp[0] = bdd_and(bddm, f->tbdd, frontier->tbdd);
    z = bdd_or(bddm, z, temp[0]);
    v_free_bdd(temp[0]);
    /* frontier = z && !oldz */
    /* Compute the successors of the frontier only. It is cheaper than
       to compute the successors of all states in z. */
    temp[0] = bdd_not(bddm, oldz);
    frontier->tbdd = bdd_and(bddm, z, temp[0]);
    if (verbose) {
      printf("EU: iteration %d, states on frontier %.0f, bdd size %.0f.\n",
             iteration++, count_bdd(frontier->tbdd),
             (double) bdd_size(bddm, frontier->tbdd, 1));
      if (verbose == 2) v_print_state("", frontier->tbdd);
    };
    v_free_bdd(temp[0]);
    if (max != -1) i--;
  };

  v_free_bdd(frontier->tbdd);
  frontier->tbdd = z;
  if (max != -1) {
    /*
     * Once we've checked inside the bounds step min times to check
     * outside the bounds.
     */
    oldz = bdd_zero(bddm);
    i = min;
    while ((oldz != frontier->tbdd) && (i > 0)) {
      v_free_bdd(oldz);
      oldz = bdd_identity(bddm, frontier->tbdd);
      /* ... also have a successor that satisfies f. */
      frontier = ex(frontier);
      temp[0] = frontier->tbdd;
      frontier->tbdd = bdd_and(bddm, f->tbdd, frontier->tbdd);
      v_free_bdd(temp[0]);
      if (verbose) {
        printf("EG: Outside bounds: iteration %d, states on frontier %.0f, bdd size %.0f.\n",
               iteration++, count_bdd(frontier->tbdd),
	       (double) bdd_size(bddm, frontier->tbdd, 1));
	if (verbose == 2) v_print_state("", frontier->tbdd);
      };
      if (max != -1) i--;
    };
  };

  v_free_bdd(oldz);
  v_free_bdd(f->tbdd);

  return(frontier);
};


/*
 * CTL_expression:
 *   A [ CTL_expression U CTL_expression ] {$$ = au($3, $5);}
 *   ;
 */
typedbdd_ptr au(f, g, min, max)
int min, max;
typedbdd_ptr f, g;
{
  bdd temp[2];
  bdd oldz;
  typedbdd_ptr z;
  int iteration = 0;
  int i;

  if (statespace == -1) fatal_error("CTL formula outside specification");
  if ((f->type != VARTYPE_BOOLEAN) || (g->type != VARTYPE_BOOLEAN)) {
    fatal_error("Non boolean expression on CTL formula: until");
  };
  if ((max != -1) && (min > max || min < 0))
      fatal_error("Invalid bounds on RTCTL formula");

  /* A[f U g] = fix z. [ g || (f && AX z)]  */

  oldz = bdd_zero(bddm);
  /* The states that satisfy A[f U g] are those that satisfy g or... */
  z = g;
  temp[0] = z->tbdd;
  z->tbdd = bdd_identity(bddm, bdd_and(bddm, z->tbdd, Reachable));
  v_free_bdd(temp[0]);
  if (verbose) {
    printf("AU: iteration %d, states %.0f, bdd size %.0f.\n",
           iteration++, count_bdd(z->tbdd),
           (double) bdd_size(bddm, z->tbdd, 1));
    if (verbose == 2) v_print_state("", z->tbdd);
  };
  if (max != -1) i = max;
            else i = min + 1;
  /*
   * - (frontier->tbdd != bdd_zero(bddm)) checks for fixpoint,
   * - (i > min) checks for RTCTL bounds (notice that in the
   *   unbounded case i is set to min+1 and it is not decremented).
   */
  while ((z->tbdd != oldz) && (i > min)) {
    v_free_bdd(oldz);
    oldz = bdd_identity(bddm, z->tbdd);
    /* ...those that satisfy f and all transitions go into states that
       satisfy A[f U g]. */
    /* z = z || (f->tbdd && ax(z)) */
    z = ax(z);
    temp[0] = bdd_and(bddm, f->tbdd, z->tbdd);
    z->tbdd = bdd_or(bddm, temp[0], oldz);
    v_free_bdd(temp[0]);
    if (verbose) {
      printf("AU: iteration %d, states %.0f, bdd size %.0f.\n",
             iteration++, count_bdd(z->tbdd),
             (double) bdd_size(bddm, z->tbdd, 1));
      if (verbose == 2) v_print_state("", z->tbdd);
    };
    if (max != -1) i--;
  };

  if (max != -1) {
    /*
     * Once we've checked inside the bounds step min times to check
     * outside the bounds.
     */
    oldz = bdd_zero(bddm);
    i = min;
    while ((oldz != z->tbdd) && (i > 0)) {
      v_free_bdd(oldz);
      oldz = bdd_identity(bddm, z->tbdd);
      z = ax(z);
      temp[0] = z->tbdd;
      z->tbdd = bdd_and(bddm, f->tbdd, z->tbdd);
      v_free_bdd(temp[0]);
      if (verbose) {
        printf("AU: Outside bounds: iteration %d, states on frontier %.0f, bdd size %.0f.\n",
	       iteration++, count_bdd(z->tbdd),
	       (double) bdd_size(bddm, z->tbdd, 1));
	if (verbose == 2) v_print_state("", z->tbdd);
      };
      if (max != -1) i--;
    };
  };

  v_free_bdd(oldz);
  v_free_bdd(f->tbdd);
  free(f);

  return(z);
};









