
/*
 * The Verus Model Checking System
 * Copyright (c) 1997 Carnegie Mellon University
 * All Rights Reserved. See file license.txt for details.
 *
 *
 * Management of the symbol table for the Verus compiler
 *
 * Sergio Campos -- 1997
 */

#include<stdio.h>
#include<strings.h>
#include<error.h>
#include<symbols.h>
#include<stdlib.h>

#include<bdd.h>

/*
 * Strings that identify symbols are stored separately. The functions below
 * implement this feature.
 */

#define STRING_BLOCK_SIZE 1024

typedef struct string_block *string_block_ptr;
struct string_block{
  char space[STRING_BLOCK_SIZE];
  int  entries;  /* number of entries in block */
  int  next;     /* position to insert next item */
};

/* global variable that points to the storage area */
string_block_ptr strings;

vresult init_string_area(strings)
string_block_ptr *strings;
{
  string_block_ptr st;

  st = (string_block_ptr) malloc(sizeof(struct string_block));
  if (st == NULL) {
    fatal_error(
          "symbols.c: init_string_area: Unable to allocate string block.");
  };
  st->entries = 0;
  st->next = 0;
  st->space[st->next] = 0;

  *strings = st;
  return(VOK);
};


/* Inserts string s in strings[position], and updates pointers. */
int insert_string(s)
char *s;
{
  int size = strlen(s) + 1;
  int position;

  if (strings->next + size > STRING_BLOCK_SIZE) {
    fatal_error("symbols.c: insert_string: String table overflow");
  };
  strcpy(&(strings->space[strings->next]), s);
  strings->entries++;
  position = strings->next;
  strings->next += size;

  return(position);
};


/* Returns a pointer to the string inserted at position. */
char *lookup_string(position)
int position;
{
  if (position < 0 || position > STRING_BLOCK_SIZE) {
    fatal_error("symbols.c: lookup_string: bad argument.");
  };
  return(&strings->space[position]);
};


/*********************************************************************/
/*
 * Data structures and functions that implement insert and lookup of
 * symbols.
 */

/* should be a prime number */
#define HASH_TABLE_SIZE 211

symbol_ptr hash_table[HASH_TABLE_SIZE];


/*
 * Hash function taken from Aho, Sethi, Ullman "Compilers --- principles,
 * techniques and tools", pp. 436. Credited to P.J. Weinberger
 */
int hash(s)
char *s;
{
  char *p;
  unsigned h = 0, g;
  for(p=s; *p != 0; p++) {
    h = ((h << 4) + (*p));
    if ((g = h&0xf0000000)) {
      h = h ^ (g >>24);
      h = h ^ g;
    };
  };
  return(h % HASH_TABLE_SIZE);
};


vresult init_symbol_table()
{
  vresult result;
  int i;

  result = init_string_area(&strings);

  function_name = malloc(MAX_SYMBOL_SIZE);
  strcpy(function_name, "");

  /* initialize hash table */
  for(i = 0; i < HASH_TABLE_SIZE; i++) {
    hash_table[i] = 0;
  };

  return(VOK);
};


/*
 * Given a symbol name, check if it is inserted properly. Returns a pointer
 * to the symbol record for that name if found, and VNOTFOUND if not.
 */
vresult lookup_symbol(name, symbolrec)
char          *name;
symbol_ptr    *symbolrec;
{
  int hashvalue = hash(name);
  symbol_ptr p, pold;

  p = hash_table[hashvalue];
  pold = p;
  while (p != 0) {
    /* check if symbol is already inserted */
    if (strcmp(p->name, name) == 0) {
      /* found it! */
      *symbolrec = p;
      return(VOK);
    };
    pold = p;
    p = p->next;
  };
  return(VNOTFOUND);
};

vresult lookup_symbol_fullname(name, symbolrec)
char          *name;
symbol_ptr    *symbolrec;
{
  char symbol_name[MAX_SYMBOL_SIZE];

  if (*function_name == 0) {
    strcpy(symbol_name, "");
  } else {
    strcpy(symbol_name, function_name);
  };
  strcat(symbol_name, name);
  return(lookup_symbol(symbol_name, symbolrec));
}


/*
 * Inserts symbol into the symbol table. If symbol was already there, simply
 * returns its position.
 */
symbol_ptr insert_symbol(name, symtype, vartype, value1, value2, value3)
char          *name;
symbol_type   symtype;
variable_type vartype;
int           value1;
int           value2;
int           value3;
{
  int hashvalue = hash(name);
  symbol_ptr p, pold;

  p = hash_table[hashvalue];
  pold = 0;
  while (p != 0) {
    /* check if symbol is already inserted */
    if (strcmp(p->name, name) == 0) {
      /* found it! */
      return(p);
    };
    pold = p;
    p = p->next;
  };

  /* Symbol was not there. Insert it. */

  p = (symbol_ptr) malloc(sizeof(struct symbol));
  p->name = lookup_string(insert_string(name));
  p->symtype = symtype;
  p->vartype = vartype;
  p->value1 = value1;
  p->value2 = value2;
  p->value3 = value3;

  p->current_value = 0;
  p->nobits = 0;

  p->next = 0;

  /* maintain a table from bdds to variable names. Used to print bdds */
  bdd_to_name_insert(p);

  if (pold == 0) {
    /* no items in this bucket, this is the first */
    hash_table[hashvalue] = p;
  } else {
    /* not the first element. Link this one to the end of the list */
    pold->next = p;
  };

  return(p);
};


#define WAIT_HASH_SIZE 211

typedef struct wait_table *wait_table_ptr;

struct wait_table{
  int            position;
  int            wc;
  wait_table_ptr next;
};

wait_table_ptr wt[WAIT_HASH_SIZE];
int wait_count;

vresult init_wait_table()
{
  int i;

  for (i = 0; i < WAIT_HASH_SIZE; i++) wt[i] = 0;
  wait_count = 0;
  return(VOK);
};

/*
 * Determines the wait count number for wait statement located at
 * position filepos in the source file.
 */
int find_wc(filepos)
int filepos;
{
  int hash_value;
  wait_table_ptr p;

  hash_value = filepos % WAIT_HASH_SIZE;
  p = wt[hash_value];
  while (p != 0) {
    if (p->position == filepos) {
      /* found it! */
      return(p->wc);
    };
    p = p->next;
  };
  /* not found */
  p = malloc(sizeof(struct wait_table));
  p->position = filepos;
  p->wc = wait_count;
  p->next = wt[hash_value];
  wt[hash_value] = p;
  wait_count++;
  return(p->wc);
};


globalvar_ptr globalvar_table = 0;

vresult insert_globalvar(gname, gi, assignedv)
char *gname;
bdd gi, assignedv;
{
  globalvar_ptr gt;

  gt = (globalvar_ptr) malloc(sizeof(struct globalvar));
  gt->gname = gname;
  gt->assignedv = assignedv;
  gt->global_identity = gi;
  gt->next = globalvar_table;
  globalvar_table = gt;

  return(VOK);
};


globalvar_ptr lookup_globalvar(gname)
char *gname;
{
  globalvar_ptr gt;

  gt = globalvar_table;
  while (gt != 0) {
    if (!strcmp(gname, gt->gname)) {
      /* found it */
      return(gt);
    };
    gt = gt->next;
  };
  fatal_error("Looking for a non existent global variable.");
  return((globalvar_ptr) 0);
};
