/*
 * The Verus Model Checking System
 * Copyright (c) 1997 Carnegie Mellon University
 * All Rights Reserved. See file license.txt for details.
 *
 *
 * Verus - A symbolic model checker for discrete time models of
 *         real-time systems.
 *
 * Sergio Campos -- 1997
 */

#include <stdio.h>
#include <sys/times.h>
#include <time.h>
#include <unistd.h>
#include <error.h>
#include <stdlib.h>
#include <symbols.h>
#include <bdduser.h>
#include <bdd.h>
#include <compile.h>
#include <modcheck.h>

#include <sys/stat.h>

extern int srcline;
extern char *yytext;

extern char *filestart;
extern long filepos;

extern int gnvars;     /* total number of boolean variables in the program */
extern int gnwaitvars; /* total number of boolean variables in wait counter*/

extern void *sbrk();

void yyerror(s)
  char *s;
{
  printf("\n%s at line %d, token %s.\n\n", s, srcline, yytext);

  exit(1);
};

int yywrap()
{
  return(1);
};

void yyparse();

char *myname;
char *filename;

char *filestart;
long filesize;
long filepos;

void open_file(filename)
char *filename;
{
  extern FILE *yyin;
  struct stat s;
  int c;

  if (stat(filename, &s) == -1) {
    fatal_error("cannot open file %s", filename);
  };
  filesize = s.st_size;
  if (!(yyin = fopen(filename, "r"))) {
    fatal_error("cannot open file %s", filename);
  };

  if (!(filestart = (char *) malloc(filesize + 1))) { 
    fatal_error("unable to allocate space for source code.");
  };

  filepos = 0;
  while ((c = fgetc(yyin)) != EOF) {
    filestart[filepos] = c;
    filepos++;
  };
  filestart[filepos] = EOF;
  filepos = 0;
  fclose(yyin);
};


bdd_manager bddm;

/*
 * Option flags
 */
int verbose = 0;       /* verbose mode can be 1 or 2. options are -v and -V. */
int reorder = 0;       /* turns on variable reordering. */
int partitionTR = 0;   /* turns on partitioned transition relations */
int short_counterex = 0; /* prints only waitcounters for counterexamples. */

char *addrstart;

extern double max_bdds_used;

extern int gc_max;

char *in_ord_file = 0;
char *out_ord_file = 0;

void print_usage(name)
char *name;
{
  printf("Usage: %s [options] <filename>, where options are:\n", name);
  printf("   -v:        Verbose level 1;\n");
  printf("   -V:        Verbose level 2;\n");
  printf("   -r:        Dynamic bdd variable reordering;\n");
  printf("   -s:        Short counterxamples - only wait counters are printed;\n");
  printf("   -l:        Uses Lazy parallel composition;\n");
  printf("   -i <file>: Read the BDD ordering from <file>;\n");
  printf("   -o <file>: Write the BDD ordering to <file>;\n");
  printf("   -c <size>: Create a bdd cache with <size> bdds (default is 500);\n");
  printf("\n\n");
};

void main(argc,argv)
int argc;
char **argv;
{
  int i;
  struct tms buffer;

  addrstart = (char *) sbrk(0);
  myname = argv[0];

  printf("\n\n*** VERUS ***\n\n");

  /* parse options. */

  i = 1;
  while (argv[i][0] == '-') {
    switch (argv[i][1]) {
      case 'v': verbose = 1; break;
      case 'V': verbose = 2; break;
      case 'r': reorder = 1; break;
      case 's': short_counterex = 1; break;
      case 'l': partitionTR = 1; break;
      case 'i': i++;
                in_ord_file = argv[i];
                break;
      case 'o': i++;
                out_ord_file = argv[i];
                break;
      case 'c': i++;
                gc_max = atoi(argv[i]);
                if (gc_max <= 0) fatal_error("Invalid cache size.");
                break;
      default: print_usage(myname);
               exit(1);
    }
    i++;
  };

  filename = argv[i];

  open_file(filename);

  init_compile();

  /* read variable ordering file */
  if (in_ord_file != 0) read_ordering_file(in_ord_file);

  /* Go for it... */
  yyparse();

  /* update variable ordering file */
  if (out_ord_file != 0) write_ordering_file(out_ord_file);

  /* make a final garbage collection */
  bdd_gc(bddm);

  printf("\n\nExecution information:\n\n");
  times(&buffer);
  printf("Time              - User               : %8g s",
	 buffer.tms_utime/(double)CLK_TCK);
  printf("  System   : %8g s\n",
	 buffer.tms_stime/(double)CLK_TCK);

  if (!partitionTR) {
    printf("BDD nodes used    - Transition relation: %10.0f",
                                            (double) bdd_size(bddm, TR, 1));
  } else {
    fn_list_ptr p;
    double total = 0;

    for(p = functions; p != 0; p = p->next) {
      total += (double) bdd_size(bddm, p->TR, 1);
    };
    printf("BDD nodes used    - Transition relation: %10.0f",
                                            (double) total);
  };
  printf("  Total    : %10.0f\n", max_bdds_used);
  printf("Bytes allocated   -                      %10.0f\n",
                                 (double)((char *)sbrk(0)-addrstart));
  printf("Boolean variables -                      %10d\n",
                                          gnwaitvars + gnvars);
  if (statespace < 1000000.0)
    printf("States            - Total              : %10.0f", statespace);
  else
    printf("States            - Total              : %10g", statespace);
  if (Reachable != bdd_zero(bddm)) {
    double r;

    r = count_bdd(Reachable);
    if (r < 1000000.0)
      printf("  Reachable: %10.0f\n", count_bdd(Reachable));
    else
      printf("  Reachable: %10g\n", count_bdd(Reachable));
  } else printf("\n");
};

